(*
Author:  Christian Kohl <christina.kohl@uibk.ac.at> (2022)
Author:  Christian Sternagel <c.sternagel@gmail.com> (2015)
Author:  Julian Nagele <julian.nagele@uibk.ac.at> (2013-2017)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2014, 2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Critical_Pair_Closure_Impl
imports
  Critical_Pairs_Impl
  Parallel_Closed
  Strongly_Closed
  Critical_Pair_Closing_Systems
  Framework.QDP_Framework_Impl
  Development_Closed
begin

definition check_strongly_closed where
  "check_strongly_closed R n = do {
     check_linear_trs R;
     check_allm (\<lambda> (b, s, t). do {
       check (\<not> Option.is_none (List.find (\<lambda>x. List.member (reachable_terms R s n) x) (reachable_terms R t (Suc 0))) \<and>
         \<not> Option.is_none (List.find (\<lambda>x. List.member (reachable_terms R t n) x) (reachable_terms R s (Suc 0))))
       (showsl_lit (STR ''the critical pair '') \<circ> showsl s \<circ> showsl_lit (STR '' <- . -> '') \<circ> showsl t \<circ>
        showsl_lit (STR '' is not strongly closed within '') \<circ> showsl n \<circ> showsl_lit (STR '' steps.''))
     }) (critical_pairs_impl R R)
     }  <+? (\<lambda>s. s \<circ> showsl_lit (STR ''\<newline>hence the following TRS is not strongly closed\<newline>'') \<circ> showsl_trs R)"

lemma check_strongly_closed:
  assumes "isOK(check_strongly_closed R n)"
  shows "CR (rstep (set R))"
proof (rule strongly_closed_linear_CR)
  from assms show "linear_trs (set R)" unfolding check_strongly_closed_def by auto
  let ?red = "\<lambda>s n. reachable_terms R s n"
  let ?opt = "\<lambda>s t. List.find (\<lambda>x. List.member (?red s n) x) (?red t (Suc 0))"
  { fix s t
    assume "\<not> Option.is_none (?opt s t)"
    then obtain r where r: "?opt s t = Some r" by (cases "?opt s t") auto
    then have "List.member (?red s n) r" unfolding find_Some_iff by auto
    then have mem:"r \<in> set (?red s n)" using in_set_member by fast
    from r have "\<exists>i<length (?red t (Suc 0)). r = (?red t (Suc 0)) ! i" unfolding find_Some_iff by auto
    then have "r \<in> set (?red t (Suc 0))" using nth_mem by blast
    with mem have "\<exists>r. (s, r) \<in> (rstep (set R))\<^sup>* \<and> (t, r) \<in> (rstep (set R))\<^sup>="
      using reachable_terms reachable_terms_one by metis
  } note sound=this
  { fix b s t
    assume "(b, s, t) \<in> critical_pairs (set R) (set R)"
    with assms have *:"\<not> Option.is_none (?opt s t) \<and> \<not> Option.is_none (?opt t s)"
      unfolding check_strongly_closed_def by fastforce
    with sound have "\<exists>r. (s, r) \<in> (rstep (set R))\<^sup>* \<and> (t, r) \<in> (rstep (set R))\<^sup>="
      and  "\<exists>u. (t, u) \<in> (rstep (set R))\<^sup>* \<and> (s, u) \<in> (rstep (set R))\<^sup>=" by blast+
  }
  then show "strongly_closed (set R)" unfolding strongly_closed_def by blast
qed

definition is_root_step :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "is_root_step R s t = (\<exists> (l, r) \<in> set R. case match s l of
      None \<Rightarrow> False
    | Some \<sigma> \<Rightarrow> r \<cdot> \<sigma> = t)"

lemma is_root_step:
  assumes "is_root_step R s t"
  shows "(s, t) \<in> rrstep (set R)"
using assms
unfolding is_root_step_def rrstep_def rstep_r_p_s_def'
by (auto split: option.splits) (force dest: match_matches)

fun is_par_rstep :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "is_par_rstep R (Var x) (Var y) = (x = y)"
| "is_par_rstep R (Fun f ss) (Fun g ts) =
    (Fun f ss = Fun g ts \<or> is_root_step R (Fun f ss) (Fun g ts) \<or>
    (if f = g \<and> length ss = length ts then list_all2 (is_par_rstep R) ss ts else False))"
| "is_par_rstep R s t = is_root_step R s t"

lemma is_par_rstep:
  assumes "is_par_rstep R s t"
  shows "(s, t) \<in> par_rstep (set R)"
using assms
proof (induct rule: is_par_rstep.induct)
  case (2 R f ss g ts)
  then consider (eq) "Fun f ss = Fun g ts"
    | (root) "is_root_step R (Fun f ss) (Fun g ts)"
    | (args) "f = g" and "length ss = length ts" and "list_all2 (is_par_rstep R) ss ts"
    by (auto split: if_splits)
  then show ?case
  proof (cases)
    case eq then show ?thesis using par_rstep_refl by auto
  next
    case root
    from is_root_step[OF this] obtain l r \<sigma>
      where "Fun f ss = l \<cdot> \<sigma>" and "Fun g ts = r \<cdot> \<sigma>" and lr:"(l, r) \<in> set R"
      unfolding rrstep_def' by auto
    then show ?thesis using par_rstep.root_step[OF lr] by metis
  next
    case args
    { fix i
      assume i:"i < length ss"
      then have si:"ss ! i \<in> set ss" by auto
      from i args(2) have ti:"ts ! i \<in> set ts" by auto
      from args(3) have "is_par_rstep R (ss ! i) (ts ! i)" using i list_all2_nthD by blast
      with 2(1)[of "ss ! i" "ts ! i"] args(1,2) si ti have "(ss ! i, ts ! i) \<in> par_rstep (set R)"
        by auto
    }
    then show ?thesis using args(1,2) by (auto)
  qed
next
  case ("3_1" R f ss y)
  then have "is_root_step R (Fun f ss) (Var y)" by auto
  from is_root_step[OF this] obtain l r \<sigma>
      where "Fun f ss = l \<cdot> \<sigma>" and "Var y = r \<cdot> \<sigma>" and lr:"(l, r) \<in> set R"
      unfolding rrstep_def' by auto
  then show ?case using par_rstep.root_step[OF lr] by metis
next
  case ("3_2" R y f ss)
  then have "is_root_step R (Var y) (Fun f ss)" by auto
  from is_root_step[OF this] obtain l r \<sigma>
      where "Fun f ss = r \<cdot> \<sigma>" and "Var y = l \<cdot> \<sigma>" and lr:"(l, r) \<in> set R"
      unfolding rrstep_def' by auto
  then show ?case using par_rstep.root_step[OF lr] by metis
qed auto

fun root_rewrite :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term list"
where
  "root_rewrite R s = concat (map (\<lambda> (l, r).
    (case match s l of
      None \<Rightarrow> []
    | Some \<sigma> \<Rightarrow> [(r \<cdot> \<sigma>)])) R)"

lemma root_rewrite_sound:
  assumes "t \<in> set (root_rewrite R s)"
  shows "(s, t) \<in> rrstep (set R)"
proof -
  from assms
  have "\<exists> l r.  (l,r) \<in> set R \<and> t \<in> set (case match s l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [r \<cdot> \<sigma>])"
    by auto
  from this obtain l r where one:
    "(l,r) \<in> set R \<and> t \<in> set (case match s l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [r \<cdot> \<sigma>])"
    by auto
  from this obtain \<sigma> where two: "match s l = Some \<sigma> \<and> t \<in> {r \<cdot> \<sigma>}" by (cases "match s l", auto)
  then have match: "l \<cdot> \<sigma> = s" using match_sound by auto
  with one match one two have "(s,t) \<in> rstep_r_p_s (set R) (l,r) [] \<sigma>" unfolding rstep_r_p_s_def by (simp add: Let_def ctxt_supt_id)
  then show "(s,t) \<in> rrstep (set R)" unfolding rstep_iff_rstep_r_p_s rrstep_def by blast
qed

fun parallel_rewrite :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term list"
where
  "parallel_rewrite R (Var x) = [Var x]"
| "parallel_rewrite R (Fun f ss) = remdups
     (root_rewrite R (Fun f ss) @ map (\<lambda>ss. Fun f ss) (product_lists (map (parallel_rewrite R) ss)))"

lemma parallel_rewrite_par_step:
  assumes "t \<in> set (parallel_rewrite R s)"
  shows "(s, t) \<in> par_rstep (set R)"
using assms
proof (induct s arbitrary: t)
  case (Fun f ss)
  then consider (root) "t \<in> set (root_rewrite R (Fun f ss))"
    | (args) "t \<in> set (map (\<lambda>ss. Fun f ss) (product_lists (map (parallel_rewrite R) ss)))"
    by force
  then show ?case
  proof (cases)
    case root
    from root_rewrite_sound[OF this] obtain l r \<sigma> where "(l, r) \<in> set R"
      and "l \<cdot> \<sigma> = Fun f ss" and "r \<cdot> \<sigma> = t"
      unfolding rrstep_def rstep_r_p_s_def by auto
    then show ?thesis using par_rstep.intros(1) by metis
  next
    case args
    then obtain ts where t:"t = Fun f ts" and ts:"ts \<in> set (product_lists (map (parallel_rewrite R) ss))"
      by auto
    then have len:"length ss = length ts" using in_set_product_lists_length by force
    { fix i
      assume i:"i < length ts"
      have "ts ! i \<in> set (parallel_rewrite R (ss ! i))"
        using ts[unfolded product_lists_set[of "_ ss"]]
        by (auto simp: list_all2_map2[of "(\<lambda>x ys. x \<in> set ys)"] intro: list_all2_nthD[OF _ i])
      with Fun(1) len i have "(ss ! i, ts ! i) \<in> par_rstep (set R)" by auto
    }
    from par_rstep.intros(2)[OF this len] show ?thesis using t by auto
  qed
qed auto

fun is_parallel_closed_cp where
  "is_parallel_closed_cp R _ (False, s, t) = is_par_rstep R t s"
| "is_parallel_closed_cp R (Some n) (True, s, t) =
    (\<not> Option.is_none (List.find (\<lambda>v. is_par_rstep R t v) (reachable_terms R s n)))"
| "is_parallel_closed_cp R None (True, s, t) =
    (\<not> Option.is_none (List.find (\<lambda>v. is_par_rstep R t v) (parallel_rewrite R s)))"

definition check_parallel_closed where
  "check_parallel_closed R n = do {
     check_left_linear_trs R;
     check_allm (\<lambda> (b, s, t). do {
       check (is_parallel_closed_cp R n (b, s, t))
         (showsl_lit (STR ''the critical pair '') \<circ> showsl s \<circ> showsl_lit (STR '' <- . -> '') \<circ> showsl t \<circ>
          showsl_lit (STR '' is not (almost) parallel closed within '') \<circ> showsl n \<circ> showsl_lit (STR '' steps.''))
     }) (critical_pairs_impl R R)
     }  <+? (\<lambda>s. s \<circ> showsl_lit (STR ''\<newline>hence the following TRS is not (almost) parallel closed\<newline>'') \<circ> showsl_trs R)"

lemma check_parallel_closed:
  assumes "isOK(check_parallel_closed R n)"
  shows "CR (rstep (set R))"
proof (rule parallel_closed_imp_CR)
  from assms show "left_linear_trs (set R)" unfolding check_parallel_closed_def by auto
  show "\<And>p q. (False, p, q) \<in> critical_pairs (set R) (set R) \<Longrightarrow> (q, p) \<in> par_rstep (set R)"
  proof -
    fix p q
    assume "(False, p, q) \<in> critical_pairs (set R) (set R)"
    with assms have "is_par_rstep R q p" unfolding check_parallel_closed_def by auto
    then show "(q, p) \<in> par_rstep (set R)" using is_par_rstep by auto
  qed
  show "\<And>p q. (True, p, q) \<in> critical_pairs (set R) (set R) \<Longrightarrow>
           \<exists>v. (q, v) \<in> par_rstep (set R) \<and> (p, v) \<in> (rstep (set R))\<^sup>*"
  proof -
    fix p q
    assume *:"(True, p, q) \<in> critical_pairs (set R) (set R)"
    show "\<exists>v. (q, v) \<in> par_rstep (set R) \<and> (p, v) \<in> (rstep (set R))\<^sup>*"
    proof (cases n)
      case (Some n)
      with assms * have "\<not> Option.is_none (List.find (\<lambda>v. is_par_rstep R q v) (reachable_terms R p n))"
        unfolding check_parallel_closed_def by auto
      then obtain v where  "List.find (\<lambda>v. is_par_rstep R q v) (reachable_terms R p n) = Some v"
        by force
      then have "is_par_rstep R q v" and "v \<in> set (reachable_terms R p n)"
        unfolding find_Some_iff by auto
      then show "\<exists>v. (q, v) \<in> par_rstep (set R) \<and> (p, v) \<in> (rstep (set R))\<^sup>*"
        using is_par_rstep reachable_terms by blast
    next
      case None
      with assms * have "\<not> Option.is_none (List.find (\<lambda>v. is_par_rstep R q v) (parallel_rewrite R p))"
        unfolding check_parallel_closed_def by auto
      then obtain v where  "List.find (\<lambda>v. is_par_rstep R q v) (parallel_rewrite R p) = Some v"
        by force
      then have "is_par_rstep R q v" and "v \<in> set (parallel_rewrite R p)"
        unfolding find_Some_iff by auto
      then show "\<exists>v. (q, v) \<in> par_rstep (set R) \<and> (p, v) \<in> (rstep (set R))\<^sup>*"
        using is_par_rstep parallel_rewrite_par_step par_rstep_rsteps by blast
    qed
  qed
qed

(*Takes a list of rewrite rules and two terms and checks for each rule whether the first term matches
the lhs and the second term the rhs of the rule. For each rule where this is the case,
returns the lists of terms corresponding to the matching substitutions. 
Terms that correspond to variables that only appear in the lhs of the rule (i.e. terms that are being
erased) don't play a role in the remainder of the computation for multisteps hence they can be ignored.*) 
fun root_steps_substs :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> (('f, 'v) term list \<times> ('f, 'v) term list) list"
where
  "root_steps_substs R s t = concat (map (\<lambda> (l, r).
    (case match s l of
      None \<Rightarrow> []
    | Some \<sigma> \<Rightarrow> (case match t r of 
        None \<Rightarrow> []
      | Some \<tau> \<Rightarrow> [(map \<sigma> (var r), map \<tau> (var r))]))) 
  R)"

lemma root_steps_substs_exists:
  assumes "(ss, ts) \<in> set (root_steps_substs R s t)"
  shows "\<exists> l r \<sigma> \<tau>. (l, r) \<in> set R \<and> l \<cdot> \<sigma> = s \<and> r \<cdot> \<tau> = t \<and> (ss, ts) = (map \<sigma> (var r), map \<tau> (var r))"
proof-
from assms obtain l r where lr:"(l,r) \<in> set R" "(ss, ts) \<in> set (case match s l of
      None \<Rightarrow> []
    | Some \<sigma> \<Rightarrow> (case match t r of 
        None \<Rightarrow> []
      | Some \<tau> \<Rightarrow> [(map \<sigma> (var r), map \<tau> (var r))]))"
  by auto
  then obtain \<sigma> where \<sigma>:"match s l = Some \<sigma>"
    by fastforce 
  with lr obtain \<tau> where \<tau>:"match t r = Some \<tau>" 
    by fastforce
  from lr \<sigma> \<tau> have "(ss, ts) = (map \<sigma> (var r), map \<tau> (var r))"
    by simp 
  with lr(1) \<sigma> \<tau> show ?thesis
    using match_matches by blast
qed

lemma size_match_subst_Fun: (*used for termination_simps*)
  assumes "is_Fun l" and "x \<in> vars_term l"
  and match:"match s l = Some \<tau>"
shows "size (\<tau> x) < size s"
proof-
  from assms(1) obtain f ts where l:"l = Fun f ts"
    by blast 
  from match have *:"l \<cdot> \<tau> = s"
    by (simp add: match_matches) 
  then obtain ss where s:"s = Fun f ss"
    unfolding l by force
  from assms(2) obtain i where i:"i < length ts" and x:"x \<in> vars_term (ts!i)" 
    unfolding l by (metis term.sel(4) var_imp_var_of_arg) 
  from * have le:"length ts = length ss" 
    unfolding s l by auto 
  moreover from * i l s have "ts!i \<cdot> \<tau> = ss!i" 
    by fastforce
  then have "size (\<tau> x) \<le> size (ss!i)"
    using match_subst_vars_term_size x by metis
  with i show ?thesis unfolding s term.size le
    by (metis add.commute add_0 add_Suc in_set_conv_nth less_Suc_eq_le size_list_estimation')
qed

context
  fixes R :: "('f, 'v) rules"
  assumes *:"list_all (\<lambda>(l, r). is_Fun l \<and> (vars_term r \<subseteq> vars_term l)) R"
begin

lemma varcond:
  "\<And>l r. (l, r) \<in> set R \<Longrightarrow> is_Fun l \<and> vars_term r \<subseteq> vars_term l"
  using * Ball_set_list_all case_prodD by (metis (no_types, lifting)) 

lemma [termination_simp]:
  assumes "(l, r) \<in> set R" 
    and "Some \<sigma> = match (Fun g ts) l" 
    and "x \<in> vars_term r"
  shows "size (\<sigma> x) < Suc (size_list size ts)"
  using assms size_match_subst_Fun varcond
  by (metis (no_types, lifting) add.right_neutral add_Suc_right subsetD term.size(4))

lemma [termination_simp]:
  assumes "(l, r) \<in> set R"
    and "Some \<tau> = match s l" 
    and "x \<in> vars_term r" 
  shows "size (\<tau> x) < size s"
  using assms size_match_subst_Fun varcond by (metis subset_iff)

lemma [termination_simp]:
  assumes "(l, r) \<in> set R"
    and "Some \<tau> = match (Var v) l" 
  shows "size (\<tau> x) = 0"
proof-
  from assms(1) varcond obtain f ss where "l = Fun f ss"
    by blast 
  with assms(2) have False
    by (metis match_matches subst_apply_eq_Var term.distinct(1)) 
  then show ?thesis 
    by simp
qed

text\<open>Given two terms s and t determine whether there exists a multi-step from s to t in the given TRS.\<close>
(*Only terminating if variable left-hand sides are excluded.*)
fun is_mstep :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  where
  "is_mstep (Fun f ss) (Fun g ts) =
    (Fun f ss = Fun g ts \<or> 
    list_ex (\<lambda> (ss, ts). list_all2 is_mstep ss ts) (root_steps_substs R (Fun f ss) (Fun g ts)) \<or> 
    (if f = g \<and> length ss = length ts then list_all2 is_mstep ss ts else False))" 
| "is_mstep s t = (s = t \<or> list_ex (\<lambda> (ss, ts). list_all2 is_mstep ss ts) (root_steps_substs R s t))"

text\<open>Show that all multisteps are covered by the definition above.\<close>
(*This is not necessary for compiling Ceta but good to know - kind of a sanity check.*)
lemma mstep_is_mstep:
  assumes "(s, t) \<in> mstep (set R)"
  shows "is_mstep s t"
  using assms proof(induct)
  case (args f n ss ts)
  then have "list_all2 is_mstep ss ts"
    by (simp add: list_all2_all_nthI) 
  with args show ?case
    by simp 
next
  case (rule l r \<sigma> \<tau>)
  from rule(1) obtain i where i:"i < length R" "R!i = (l, r)"
    by (metis in_set_idx) 
  obtain \<sigma>' where sigma:"match (l \<cdot> \<sigma>) l = Some \<sigma>'" "(\<forall>x\<in>vars_term l. \<sigma> x = \<sigma>' x)" 
    by (meson match_complete')
  obtain \<tau>' where tau:"match (r \<cdot> \<tau>) r = Some \<tau>'" "(\<forall>x\<in>vars_term r. \<tau> x = \<tau>' x)" 
    by (meson match_complete')
  let ?matches="(map (\<lambda>(l', r'). case match (l \<cdot> \<sigma>) l' of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> case match (r \<cdot> \<tau>) r' of None \<Rightarrow> [] | Some \<tau> \<Rightarrow> [(map \<sigma> (var r'), map \<tau> (var r'))]) R)"
  have "i < length ?matches"
    using i(1) by auto 
  moreover have "(map \<sigma>' (var r), map \<tau>' (var r)) \<in> set (?matches ! i)"
    using sigma(1) tau(1) i by simp
  ultimately have "(map \<sigma>' (var r), map \<tau>' (var r)) \<in> set (root_steps_substs R (l\<cdot>\<sigma>) (r\<cdot>\<tau>))"
    unfolding root_steps_substs.simps by (metis (no_types, lifting) concat_nth concat_nth_length in_set_conv_nth) 
  then obtain j where j:"j < length (root_steps_substs R (l\<cdot>\<sigma>) (r\<cdot>\<tau>))" "root_steps_substs R (l\<cdot>\<sigma>) (r\<cdot>\<tau>) ! j = (map \<sigma>' (var r), map \<tau>' (var r))"
    by (metis in_set_idx)
  have "(\<lambda> (ss, ts). list_all2 is_mstep ss ts) ((root_steps_substs R (l\<cdot>\<sigma>) (r\<cdot>\<tau>))!j)"
  proof-
    {fix i assume "i < length (var r)"
      then have *:"var r!i \<in> vars_term r"
        by (metis comp_apply nth_mem set_remdups set_rev set_vars_term_list) 
      then have "var r!i \<in> vars_term l" 
        using varcond rule.hyps(1) by blast 
      moreover then have "\<sigma>' (var r !i) = \<sigma> (var r ! i)" 
        using sigma(2) * by simp
      moreover have "\<tau>' (var r !i) = \<tau> (var r ! i)"
        using * tau(2) by simp
      ultimately have "is_mstep (\<sigma>' (var r !i)) (\<tau>' (var r!i))" 
        using rule(2) by force
    }
    then have "list_all2 is_mstep (map \<sigma>' (var r)) (map \<tau>' (var r))"
      by (simp add: list_all2_conv_all_nth) 
    then show ?thesis 
      unfolding j(2) by fastforce 
  qed
  then have *:"list_ex (\<lambda> (ss, ts). list_all2 is_mstep ss ts) (root_steps_substs R (l\<cdot>\<sigma>) (r\<cdot>\<tau>))"
    using j by (meson list_ex_length) 
  then show ?case proof(cases "l \<cdot> \<sigma>")
    case (Var x)
    then show ?thesis 
      using * is_mstep.simps(2) by presburger 
  next
    case (Fun f ts)
    note f=this
    show ?thesis proof(cases "r \<cdot> \<tau>")
      case (Var x)
      then show ?thesis 
        using * is_mstep.simps(3) by presburger 
    next
      case (Fun g ss)
      then show ?thesis 
        using f * is_mstep.simps(1) by presburger 
    qed
  qed 
qed simp

lemma is_mstep:
  assumes "is_mstep s t"
  shows "(s, t) \<in> mstep (set R)"
using assms proof (induct rule: is_mstep.induct)
  case (1 f ss g ts)
  then consider (eq) "Fun f ss = Fun g ts"
    | (root) "list_ex (\<lambda> (ss, ts). list_all2 is_mstep ss ts) (root_steps_substs R (Fun f ss) (Fun g ts))"
    | (args) "f = g" and "length ss = length ts" and "list_all2 is_mstep ss ts"
    by (auto split: if_splits)
  then show ?case proof(cases)
    case root
    then obtain i where "i < length (root_steps_substs R (Fun f ss) (Fun g ts))" "(\<lambda> (ss, ts). list_all2 is_mstep ss ts) ((root_steps_substs R (Fun f ss) (Fun g ts))!i)"
      using list_ex_length by blast
    then obtain ss' ts' where ss'ts':"(ss' , ts') \<in> set (root_steps_substs R (Fun f ss) (Fun g ts))" "list_all2 is_mstep ss' ts'"
      using nth_mem by fastforce 
    with root_steps_substs_exists obtain l r \<sigma> \<tau> where lr:"(l, r) \<in> set R" 
      and l:"l \<cdot> \<sigma> = Fun f ss" and r:"r \<cdot> \<tau> = Fun g ts" 
      and \<sigma>\<tau>:"(ss', ts') = (map \<sigma> (var r), map \<tau> (var r))"
      by blast
    let ?\<tau>="\<lambda>x. (if x \<in> vars_term r then \<tau> x else \<sigma> x)"
    from r have r':"r \<cdot> ?\<tau> = Fun g ts"
      by (smt (verit, del_insts) term_subst_eq) 
    { fix x assume "x \<in> vars_term l"
      then have "(\<sigma> x, ?\<tau> x) \<in> mstep (set R)" proof(cases "x \<in> vars_term r")
        case True
        then obtain i where i:"i < length (var r)" "var r ! i = x"
          by (metis in_set_idx set_vars_term_list vars_term_list_var)  
        then have i1:"i < length ss'" 
          using \<sigma>\<tau> by simp 
        from i have i2:"i < length ts'"
          using \<sigma>\<tau> by simp
        from ss'ts'(2) i1 i2 have "is_mstep (ss' ! i) (ts' ! i)"
          using list_all2_nthD by blast
        with 1(1)[OF ss'ts'(1)] i1 i2 have "(ss' ! i, ts' ! i) \<in> mstep (set R)"
          by auto
        then show ?thesis 
          using i \<sigma>\<tau> by auto 
      qed simp
    }
    then show ?thesis 
      using mstep.rule[OF lr] l r' by force 
  next
    case args
    { fix i
      assume i:"i < length ss"
      then have si:"ss ! i \<in> set ss" by auto
      from i args(2) have ti:"ts ! i \<in> set ts" by auto
      from args(3) have "is_mstep (ss ! i) (ts ! i)" using i list_all2_nthD by blast
      with 1(2)[of "ss ! i" "ts ! i"] args(1,2) si ti have "(ss ! i, ts ! i) \<in> mstep (set R)"
        by auto
    }
    then show ?thesis using args(1,2)
      by (simp add: mstep.args)
  qed simp
next
  case ("2_1" v t) 
  then consider (eq) "Var v = t"
    | (root) "list_ex (\<lambda> (ss, ts). list_all2 is_mstep ss ts) (root_steps_substs R (Var v) t)"
    by auto
  then show ?case proof(cases)
    case root
    then obtain i where "i < length (root_steps_substs R (Var v) t)" "(\<lambda> (ss, ts). list_all2 is_mstep ss ts) ((root_steps_substs R (Var v) t)!i)"
      unfolding is_mstep.simps using list_ex_length by blast
    then obtain ss' ts' where ss'ts':"(ss' , ts') \<in> set (root_steps_substs R (Var v) t)" "list_all2 is_mstep ss' ts'"
      using nth_mem by fastforce 
    with root_steps_substs_exists obtain l r \<sigma> \<tau> where lr:"(l, r) \<in> set R" 
      and l:"l \<cdot> \<sigma> = Var v" and r:"r \<cdot> \<tau> = t" 
      and \<sigma>\<tau>:"(ss', ts') = (map \<sigma> (var r), map \<tau> (var r))"
      by blast
    let ?\<tau>="\<lambda>x. (if x \<in> vars_term r then \<tau> x else \<sigma> x)"
    from r have r':"r \<cdot> ?\<tau> = t"
      by (smt (verit, del_insts) term_subst_eq) 
    { fix x assume "x \<in> vars_term l"
      then have "(\<sigma> x, ?\<tau> x) \<in> mstep (set R)" proof(cases "x \<in> vars_term r")
        case True
        then obtain i where i:"i < length (var r)" "var r ! i = x"
          by (metis in_set_idx set_vars_term_list vars_term_list_var)  
        then have i1:"i < length ss'" 
          using \<sigma>\<tau>  by simp 
        from i have i2:"i < length ts'"
          using \<sigma>\<tau> by simp
        from ss'ts'(2) i1 i2 have "is_mstep (ss' ! i) (ts' ! i)"
          using list_all2_nthD by blast
        with "2_1"(1)[OF ss'ts'(1)] i1 i2 have "(ss' ! i, ts' ! i) \<in> mstep (set R)"
          by auto
        then show ?thesis
          using i \<sigma>\<tau> by auto 
      qed simp
    }
    then show ?thesis 
      using mstep.rule[OF lr] l r' by force 
  qed simp
next
  case ("2_2" s v)
  then consider (eq) "s = Var v"
    | (root) "list_ex (\<lambda> (ss, ts). list_all2 is_mstep ss ts) (root_steps_substs R s (Var v))"
    by auto
  then show ?case proof(cases)
    case root
    then obtain i where "i < length (root_steps_substs R s (Var v))" "(\<lambda> (ss, ts). list_all2 is_mstep ss ts) ((root_steps_substs R s (Var v))!i)"
      unfolding is_mstep.simps using list_ex_length by blast
    then obtain ss' ts' where ss'ts':"(ss' , ts') \<in> set (root_steps_substs R s (Var v))" "list_all2 is_mstep ss' ts'"
      using nth_mem by fastforce 
    with root_steps_substs_exists obtain l r \<sigma> \<tau> where lr:"(l, r) \<in> set R" 
      and l:"l \<cdot> \<sigma> = s" and r:"r \<cdot> \<tau> = Var v" 
      and \<sigma>\<tau>:"(ss', ts') = (map \<sigma> (var r), map \<tau> (var r))"
      by blast
    let ?\<tau>="\<lambda>x. (if x \<in> vars_term r then \<tau> x else \<sigma> x)"
    from r have r':"r \<cdot> ?\<tau> = Var v"
      by (smt (verit, del_insts) term_subst_eq) 
    { fix x assume "x \<in> vars_term l"
      then have "(\<sigma> x, ?\<tau> x) \<in> mstep (set R)" proof(cases "x \<in> vars_term r")
        case True
        then obtain i where i:"i < length (var r)" "var r ! i = x"
          by (metis in_set_idx set_vars_term_list vars_term_list_var)  
        then have i1:"i < length ss'" 
          using \<sigma>\<tau>  by simp 
        from i have i2:"i < length ts'"
          using \<sigma>\<tau> by simp
        from ss'ts'(2) i1 i2 have "is_mstep (ss' ! i) (ts' ! i)"
          using list_all2_nthD by blast
        with "2_2"(1)[OF ss'ts'(1)] i1 i2 have "(ss' ! i, ts' ! i) \<in> mstep (set R)"
          by auto
        then show ?thesis
          using i \<sigma>\<tau> by auto 
      qed simp
    }
    then show ?thesis 
      using mstep.rule[OF lr] l r' by force 
  qed simp
qed

(*super simple definition since we are not considering almost development closed yet, so no distinction for root steps*)
fun is_development_closed_cp where
  "is_development_closed_cp s t = is_mstep t s"

end

text\<open>We need to be able to export code for @{const is_mstep}, hence the following definitions
(adapted from template by Rene).\<close>
(*New type for well-formed TRSs*)
typedef ('f, 'v) wfTRS = "{R :: ('f, 'v) rules. list_all (\<lambda>(l,r). is_Fun l \<and> vars_term r \<subseteq> vars_term l) R}" 
proof-
  have "list_all (\<lambda>(l,r). is_Fun l) [(Fun ''c'' [], Var ''x'')]" (*had to be strangely specific*)
    by simp
  then show ?thesis
    using list_all_simps(2) by blast 
qed

setup_lifting type_definition_wfTRS

lift_definition get_TRS :: "('f, 'v) wfTRS \<Rightarrow> ('f, 'v) rules" is "\<lambda> R. R" .

lemma is_wf_get_TRS: "list_all (\<lambda>(l,r). is_Fun l  \<and> vars_term r \<subseteq> vars_term l) (get_TRS R')" 
  by (transfer, auto)

definition "is_mstep' R s t = is_mstep (get_TRS R) s t" 

lemmas is_mstep'_simps = is_mstep.simps[OF is_wf_get_TRS, folded is_mstep'_def]

declare is_mstep'_simps[code]

definition "is_development_closed_cp' R s t = is_development_closed_cp (get_TRS R) s t" 

lemmas is_development_closed_cp'_simps = is_development_closed_cp.simps[OF is_wf_get_TRS, folded is_development_closed_cp'_def is_mstep'_def]

declare is_development_closed_cp'_simps[code]

lift_definition (code_dt) get_wfTRS :: "('f, 'v) rules \<Rightarrow> ('f, 'v) wfTRS option" is 
  "\<lambda> R. if list_all (\<lambda>(l,r). is_Fun l \<and> vars_term r \<subseteq> vars_term l) R then Some R else None"
  by auto

definition err_wf where "err_wf = STR ''TRS is not well-formed''" 
(*should actually never be printed, since TRS is checked before even calling is_mstep*) 

lemma [code]: "is_mstep R = (case get_wfTRS R of Some R' \<Rightarrow> is_mstep' R' 
  | None \<Rightarrow> Code.abort err_wf (\<lambda> _. is_mstep R ))" 
proof (cases "list_all (\<lambda>(l,r). is_Fun l \<and> vars_term r \<subseteq> vars_term l) R")
  case False
  hence "get_wfTRS R = None" by transfer auto
  thus ?thesis by auto
next
  case True
  hence "\<exists> R'. get_wfTRS R = Some R' \<and> get_TRS R' = R"
    by (transfer, auto)
  thus ?thesis unfolding is_mstep'_def by auto
qed

lemma [code]: "is_development_closed_cp R = (case get_wfTRS R of Some R' \<Rightarrow> is_development_closed_cp' R' 
  | None \<Rightarrow> Code.abort err_wf (\<lambda> _. is_development_closed_cp R ))" 
proof (cases "list_all (\<lambda>(l,r). is_Fun l \<and> vars_term r \<subseteq> vars_term l) R")
  case False
  hence "get_wfTRS R = None" by transfer auto
  thus ?thesis by auto
next
  case True
  hence "\<exists> R'. get_wfTRS R = Some R' \<and> get_TRS R' = R"
    by (transfer, auto)
  thus ?thesis unfolding is_development_closed_cp'_def by auto
qed


definition check_development_closed where
  "check_development_closed R = do {
     check_wf_trs R;
     check_left_linear_trs R;
     check_allm (\<lambda> (b, s, t). do {
       check (is_development_closed_cp R s t)
         (showsl_lit (STR ''the critical pair '') \<circ> showsl s \<circ> showsl_lit (STR '' <- . -> '') \<circ> showsl t \<circ>
          showsl_lit (STR '' is not development closed''))
     }) (critical_pairs_impl R R)
     }  <+? (\<lambda>s. s \<circ> showsl_lit (STR ''\<newline>hence the following TRS is not development closed\<newline>'') \<circ> showsl_trs R)"

lemma check_development_closed:
  assumes "isOK(check_development_closed R)"
  shows "CR (rstep (set R))"
proof (rule mstep_closed_imp_CR)
  from assms show "left_linear_trs (set R)" 
    unfolding check_development_closed_def by auto 
  from assms have varcond:"list_all (\<lambda>(l,r). is_Fun l \<and> vars_term r \<subseteq> vars_term l) R"
    unfolding check_development_closed_def check_wf_trs_def check_varcond_no_Var_lhs list_all_iff by auto  
  then show "\<And>l r. (l, r) \<in> set R \<Longrightarrow> is_Fun l \<and> vars_term r \<subseteq> vars_term l"
    using Critical_Pair_Closure_Impl.varcond by blast 
  show "\<And>b p q. (b, p, q) \<in> critical_pairs (set R) (set R) \<Longrightarrow> (q, p) \<in> mstep (set R)"
  proof -
    fix b p q
    assume "(b, p, q) \<in> critical_pairs (set R) (set R)"
    with assms have "is_development_closed_cp R p q" 
      unfolding check_development_closed_def by auto
    then have "is_mstep R q p" 
      using varcond is_development_closed_cp.simps by blast
    then show "(q, p) \<in> mstep (set R)" using is_mstep[OF varcond] by auto
  qed
qed

fun is_critical_pair_closing_cp where
  "is_critical_pair_closing_cp C n (False, s, t) =
    (\<not> Option.is_none (List.find (\<lambda>v. is_par_rstep C t v) (reachable_terms C s n)))"
| "is_critical_pair_closing_cp C n (True, s, t) =
    (\<not> Option.is_none (List.find (\<lambda>x. List.member (reachable_terms C s n) x) (reachable_terms C t n)))"

definition check_critical_pair_closing :: "('f:: showl, string) rules \<Rightarrow> ('f, string) rules \<Rightarrow> nat \<Rightarrow> showsl check"
where
  "check_critical_pair_closing R C n = do {
     check_left_linear_trs R;
     check_subseteq C R <+? (\<lambda>t. showsl_lit (STR ''C not a subsystem of R''));
     check_allm (\<lambda> (b, s, t). do {
       check (is_critical_pair_closing_cp C n (b, s, t))
         (showsl_lit (STR ''the critical pair '') \<circ> showsl s \<circ> showsl_lit (STR '' <- . -> '') \<circ> showsl t \<circ>
          showsl_lit (STR '' is not closed within '') \<circ> showsl n \<circ> showsl_lit (STR '' steps.''))
     }) (critical_pairs_impl R R)
     }  <+? (\<lambda>s. s \<circ> showsl_lit (STR ''\<newline>hence the following TRS is not critical pair closing\<newline>'') \<circ> showsl_trs R)"

lemma check_critical_pair_closing:
  assumes "isOK(check_critical_pair_closing R C n)"
    "SN (rstep (set C))"
  shows "CR (rstep (set R))"
proof (rule cpcs_sn)
  from assms show "left_linear_trs (set R)" unfolding check_critical_pair_closing_def by auto
  from assms show "SN (rstep (set C))" ..
  from assms[unfolded check_critical_pair_closing_def,simplified]
  have closed:"\<And> b p q. (b, p, q) \<in> critical_pairs (set R) (set R) \<Longrightarrow> is_critical_pair_closing_cp C n (b, p, q)"
    by fast
  show "critical_pair_closing (set C) (set R)" unfolding critical_pair_closing_def
  proof
    from assms show "set C \<subseteq> set R"  unfolding check_critical_pair_closing_def by auto
    { fix b p q
      assume cp:"(b, p, q) \<in> critical_pairs (set R) (set R)"
      have "(p, q) \<in> (rstep (set C))\<^sup>\<down>"
      proof (cases b)
        case True
        with closed[of b p q] cp
        have "\<not> Option.is_none (find (List.member (reachable_terms C p n)) (reachable_terms C q n))" by auto
        then obtain v where  "find (List.member (reachable_terms C p n)) (reachable_terms C q n) = Some v"
          by force
        then have "v \<in> set (reachable_terms C p n) \<and> v \<in> set (reachable_terms C q n)"
          unfolding find_Some_iff by (meson in_set_member nth_mem)
        then show ?thesis by (auto dest: reachable_terms)
      next
        case False
        with closed[of b p q] cp
        have "\<not> Option.is_none (find (is_par_rstep C q) (reachable_terms C p n))" by auto
        then obtain v where  "find (is_par_rstep C q) (reachable_terms C p n) = Some v"
          by force
        then have "is_par_rstep C q v" and "v \<in> set (reachable_terms C p n)"
          unfolding find_Some_iff by auto
        then show ?thesis using is_par_rstep reachable_terms par_rstep_rsteps by blast
      qed
    }
    then show "\<forall>(b, p, q) \<in> critical_pairs (set R) (set R). (p, q) \<in> (rstep (set C))\<^sup>\<down>" by auto
  qed
  show "\<And>p q. (False, p, q) \<in> critical_pairs (set R) (set R) \<Longrightarrow>
           \<exists>v. (q, v) \<in> par_rstep (set C) \<and> (p, v) \<in> (rstep (set C))\<^sup>*"
  proof -
    fix p q
    assume "(False, p, q) \<in> critical_pairs (set R) (set R)"
    with closed[of False p q]
    have "\<not> Option.is_none (find (is_par_rstep C q) (reachable_terms C p n))" by auto
    then obtain v where  "find (is_par_rstep C q) (reachable_terms C p n) = Some v"
      by force
    then have "is_par_rstep C q v" and "v \<in> set (reachable_terms C p n)"
      unfolding find_Some_iff by auto
    then show "\<exists>v. (q, v) \<in> par_rstep (set C) \<and> (p, v) \<in> (rstep (set C))\<^sup>*"
      using is_par_rstep reachable_terms par_rstep_rsteps by blast
  qed
qed

end
