(*
Author:  Christina Kohl <christina.kohl@uibk.ac.at> (2022)
License: LGPL (see file COPYING.LESSER)
*)
theory Utils
imports 
  First_Order_Terms.Matching
  TRS.Term_Impl
begin

section\<open>arbitrary results\<close>
lemma map_of_contains:
  assumes "(a, b) \<in> set xs" "\<forall>(a', b') \<in> set xs. (a' = a \<longrightarrow> b' = b)"
  shows "(map_of xs) a = Some b"
  using assms(1) assms(2) map_of_SomeD weak_map_of_SomeI by force

lemma remdups_no_duplicates:
  assumes "\<forall>x. length (filter ((=) x) xs) \<le> 1"
  shows "remdups xs = xs"
  using assms proof(induct xs)
  case (Cons x xs)
  then have IH:"remdups xs = xs"
    by (metis filter.simps(2) impossible_Cons le_cases le_trans) 
  from Cons(2) have "length (filter ((=) x) xs) = 0"
    by (metis (mono_tags) One_nat_def add.right_neutral add_Suc_right filter.simps(2) le_less length_0_conv less_Suc0 list.simps(3) list.size(4) nat.inject)
  then have "x \<notin> set (xs)"
    by (metis filter_empty_conv length_0_conv) 
  with IH show ?case
    by simp
qed simp

lemma obtain_list_from_property:
  assumes "\<forall>x \<in> set xs. \<exists>a. P a x"
  shows "\<exists>as. length as = length xs \<and> (\<forall>i < length xs. P (as!i) (xs!i))"
  using assms proof(induct xs)
  case (Cons a xs)
  then show ?case
    by (metis length_map nth_map nth_mem)
qed simp

lemma zip_symm: "(x, y) \<in> set (zip xs ys) \<Longrightarrow> (y, x) \<in> set (zip ys xs)"
  by (induct xs ys rule:list_induct2') auto

lemma sum_list_elem:
  "(\<Sum>x\<leftarrow>[y]. f x) = f y"
  by simp

lemma card_Union_Sum:
  assumes "is_partition (map f [0..<length xs])" 
    and "\<forall>i < length xs. finite (f i)"
  shows "card (\<Union>i<length xs. f i) = (\<Sum>i<length xs. card (f i))"
proof-
  from assms(1) have disj:"pairwise (\<lambda>s t. disjnt (f s) (f t)) {..<length xs}" 
    unfolding pairwise_def is_partition_alt is_partition_alt_def disjnt_def by simp
  then have "pairwise disjnt (f ` {..<length xs})" 
    by (metis (mono_tags, lifting) pairwiseD pairwise_imageI) 
  then have "card (\<Union>i<length xs. f i) = sum card (f ` {..<length xs})" 
    using assms(2) card_Union_disjoint by (metis (mono_tags, lifting) imageE lessThan_iff) 
  with disj show ?thesis 
    using sum_card_image by (metis finite_lessThan)
qed

lemma sum_sum_concat: "(\<Sum>i<length xs. \<Sum>x\<leftarrow>f (xs!i). g x) = (\<Sum>x\<leftarrow>concat (map f xs). g x)"
proof(induct xs)
  case (Cons a xs)
  then show ?case unfolding list.map concat.simps map_append sum_list_append
    by (metis (mono_tags, lifting) length_nth_simps(2) nth_Cons_0 nth_Cons_Suc sum.cong sum.lessThan_Suc_shift)
qed simp

lemma sum_list_zero:
  assumes "\<forall>i < length xs. f (xs!i) = 0"
  shows "(\<Sum>x\<leftarrow>xs. f x) = 0"
  by (metis assms map_eq_conv' monoid_add_class.sum_list_0)

lemma concat_map2_zip:
  assumes "length xs = length ys" 
    and "\<forall>i < length xs. length (xs!i) = length (ys!i)"
  shows "concat (map2 zip xs ys) = zip (concat xs) (concat ys)"
  using assms proof(induct xs arbitrary:ys rule:rev_induct)
  case (snoc x xs)
  from snoc(2) obtain y ys' where y:"ys = ys'@[y]"
    by (metis append_is_Nil_conv length_0_conv neq_Nil_conv rev_exhaust) 
  moreover with snoc(2) have l:"length xs = length ys'" by simp
  moreover with snoc(3) have l':"\<forall>i < length xs. length (xs!i) = length (ys'!i)"
    unfolding y by (metis (no_types, lifting) Ex_less_Suc add_Suc_right append.right_neutral append_Cons_nth_left length_Cons length_append order_less_trans) 
  ultimately have IH:"concat (map2 zip xs ys') = zip (concat xs) (concat ys')"
    using snoc(1) by presburger
  have *:"concat (map2 zip (xs @ [x]) ys) = concat (map2 zip xs ys') @ (zip x y)" 
    unfolding y zip_append[OF l] by simp
  have "length (concat xs) = length (concat ys')" 
    using l l' eq_length_concat_nth by blast 
  then show ?case 
    unfolding * IH unfolding y concat_append using zip_append by simp   
qed simp

section\<open>Take and drop.\<close>
lemma take_equal_xs_ys:
  assumes "i < length xs" "length xs = length ys"
    and "\<forall>j < length xs. j \<noteq> i \<longrightarrow> xs!j = ys!j"
  shows "take i xs = take i ys"
  using assms by (smt (verit, del_insts) min.absorb4 min.strict_boundedE nat_less_le nth_take_lemma) 

lemma drop_equal_xs_ys:
  assumes "i < length xs" "length xs = length ys"
    and "\<forall>j < length xs. j \<noteq> i \<longrightarrow> xs!j = ys!j"
  shows "drop (Suc i) xs = drop (Suc i) ys"
  using assms proof(induct xs arbitrary:ys i)
  case (Cons x xs')
  from Cons(3) obtain y ys' where y:"ys = y # ys'"
    by (metis length_0_conv list.exhaust) 
  then show ?case proof(cases i)
    case 0
    with y Cons(4) have "xs' = ys'" 
      unfolding 0 by (metis Cons.prems(2) Suc_inject Suc_mono Zero_not_Suc length_Cons length_nth_simps(4) nth_equalityI) 
    then show ?thesis unfolding 0 y drop_Suc_Cons drop_0 by simp
  next
    case (Suc n)
    with Cons have "drop (Suc n) xs' = drop (Suc n) ys'" 
      unfolding y by (metis Suc_inject Suc_less_eq length_nth_simps(2) nth_Cons_Suc)
    then show ?thesis unfolding Suc y drop_Suc_Cons by simp
  qed
qed simp

lemma union_take_drop_list:
  assumes "i < length xs"
  shows "(set (take i xs)) \<union> (set (drop (Suc i) xs)) = {xs!j | j. j < length xs \<and> j \<noteq> i}"
proof-
  from assms have i:"i \<le> length xs" by simp
  have set1:"set (take i xs) = {xs ! j |j. j < i}" 
    using nth_image[OF i] unfolding image_def by fastforce
  from assms have i:"Suc i \<le> length xs" by simp
  have set2:"set (drop (Suc i) xs) = {xs !j |j. i < j \<and> j < length xs}" proof
    {fix x assume "x \<in> set (drop (Suc i) xs)" 
      then have "x \<in> {xs !j |j. i < j \<and> j < length xs}" 
        unfolding set_conv_nth nth_drop[OF i] length_drop by auto 
    }
    then show "set (drop (Suc i) xs) \<subseteq> {xs ! j |j. i < j \<and> j < length xs}" by auto
    {fix x assume "x \<in> {xs !j |j. i < j \<and> j < length xs}" 
      then have "x \<in> set (drop (Suc i) xs)" 
        unfolding set_conv_nth nth_drop[OF i] length_drop
        by (smt (verit, best) Suc_leI add_diff_inverse_nat i mem_Collect_eq nat_add_left_cancel_less not_less_eq_eq)
    }
    then show "{xs ! j |j. i < j \<and> j < length xs} \<subseteq> set (drop (Suc i) xs)" by auto
  qed
  {fix x assume "x \<in> set (take i xs) \<union> set (drop (Suc i) xs)"
    then consider "x \<in> set (take i xs)" | "x \<in> set (drop (Suc i) xs)" by blast
    then have "x \<in> {xs ! j |j. j < length xs \<and> j \<noteq> i}" proof(cases)
      case 1
      with set1 show ?thesis using in_set_idx by fastforce
    next
      case 2
      with set2 show ?thesis using in_set_idx by fastforce
    qed
  }moreover
  {fix x assume "x \<in> {xs ! j |j. j < length xs \<and> j \<noteq> i}"
    then obtain j where "x = xs!j" and j:"j < length xs" "j \<noteq> i"
      by blast
    then have "x \<in> set (take i xs) \<union> set (drop (Suc i) xs)" 
      using set1 set2 using nat_neq_iff by auto
  }
  ultimately show ?thesis by auto
qed

section\<open>Facts about variables and positions\<close>
text\<open>Duplicate free version of @{const vars_term_list} that preserves order of appearance of variables.\<close>
(*Looks complicated because of the way remdups is defined.*)
abbreviation var :: "('f, 'v) term \<Rightarrow> 'v list" where "var t \<equiv> (rev \<circ> remdups \<circ> rev) (vars_term_list t)"

abbreviation varposs_list :: "('f, 'v) term \<Rightarrow> pos list" 
  where "varposs_list t \<equiv> filter (\<lambda>p. p \<in> varposs t) (poss_list t)" 

lemma single_var[simp]: "var (Var x) = [x]"
  by (simp add: vars_term_list.simps(1)) 

lemma args_poss: 
  assumes "i # p \<in> poss t"
  shows "\<exists>f ts. t = Fun f ts \<and> p \<in> poss (ts!i) \<and> i < length ts"
  by (metis Cons_poss_Var assms poss.elims poss_Cons_poss term.sel(4))

lemma vars_term_list_var:
  assumes "i < length (vars_term_list t)"
  shows "\<exists>j < length (var t). (vars_term_list t)!i = (var t)!j"
  by (metis assms in_set_idx nth_mem o_apply set_remdups set_rev)

lemma varposs_in_varposs_list:
  assumes "p \<in> varposs t"
  shows "\<exists>i < length (varposs_list t). (varposs_list t)!i = p"
  by (metis (mono_tags, lifting) assms in_set_conv_nth mem_Collect_eq poss_list_sound set_filter varposs_imp_poss)

lemma varposs_set_varposs_list: "varposs t = set (varposs_list t)"
proof
  show "varposs t \<subseteq> set (varposs_list t)"
    by (simp add: subsetI varposs_imp_poss) 
  show "set (varposs_list t) \<subseteq> varposs t" 
    using varposs_in_varposs_list by auto 
qed

lemma varposs_list_fun:
  "varposs_list (Fun f ss) = concat (map (\<lambda> (i, ps). map ((#) i) ps) (zip [0 ..< length ss] (map varposs_list ss)))" 
  (is "?varposs = concat (map ?g ?zip)")
proof-
  let ?P="(\<lambda>p. p \<in> varposs (Fun f ss))" 
  let ?zip'="zip [0..<length ss] (map poss_list ss)"
  have l1:"length ?zip = length ss" by simp  
  have l2:"length ?zip' = length ss" by simp 
  {fix i assume i:"i < length ss"
    then have *:"?g (?zip'!i) = map ((#) i) (poss_list (ss!i))"
      by simp  
    have "\<forall>p \<in> set (poss_list (ss!i)). (i#p \<in> varposs (Fun f ss)) = (p \<in> varposs (ss!i))"
      using i by auto
    then have "((filter ?P) \<circ> ?g) (?zip'!i) = map ((#) i) (varposs_list (ss ! i))" 
      unfolding o_apply * filter_map  by (metis (no_types, lifting) filter_cong)
    then have "((filter ?P) \<circ> ?g) (?zip'!i) = ?g (?zip!i)"
      using i by auto
  }
  with l1 l2 have "(map ((filter ?P) \<circ> ?g) (zip [0..<length ss] (map poss_list ss))) = (map ?g ?zip)"
    by (simp add: map_nth_eq_conv)   
  then show ?thesis 
    by (simp add: filter_concat poss_list.simps(2))   
qed

lemma varposs_list_fun_elem:
  assumes "i < length (varposs_list (Fun f ts))" 
    and "varposs_list (Fun f ts) ! i = j#r"
  shows "j < length ts \<and> (\<exists>i'. i' < length (varposs_list (ts!j)) \<and> varposs_list (ts!j) ! i' = r)"
proof-
  from assms have j:"j < length ts" 
    by (metis (no_types, lifting) mem_Collect_eq nth_mem poss_Cons_poss poss_list_sound set_filter term.sel(4)) 
  with assms have "r \<in> varposs (ts!j)"
    using nth_mem by fastforce 
  then have "(\<exists>i'. i' < length (varposs_list (ts!j)) \<and> varposs_list (ts!j) ! i' = r)"
    using varposs_in_varposs_list by blast 
  with j show ?thesis by simp
qed

lemma length_varposs_list: "length (varposs_list t) = length (vars_term_list t)"
proof(induct t)
  case (Var x)
  then show ?case unfolding varposs.simps vars_term_list.simps poss_list.simps by simp
next
  case (Fun f ts)
  let ?xs="map2 (\<lambda>x. map ((#) x)) [0..<length ts] (map varposs_list ts)"
  let ?ys="map vars_term_list ts"
  have l1:"length ?xs = length ts"
    by simp  
  have l2:"length ?ys = length ts" 
    by simp
  {fix i assume i:"i < length ts"
    then have "(zip [0..<length ts] (map varposs_list ts)) ! i = (i, varposs_list (ts!i))" by simp 
    with i have "?xs!i = (map ((#) i)) (varposs_list (ts!i))" by simp
    then have "length (?xs!i) = length (varposs_list (ts!i))" by simp 
    with i have "length (?xs!i) = length (?ys!i)" using Fun.hyps by simp
  }
  then show ?case 
    unfolding varposs_list_fun vars_term_list.simps using eq_length_concat_nth[of ?xs ?ys] l1 l2 by presburger  
qed

lemma vars_term_list_varposs_list:
  assumes "i < length (vars_term_list t)" 
  shows "Var ((vars_term_list t)!i) = t|_((varposs_list t)!i)"
  using assms proof(induct t arbitrary:i)
  case (Var x)
  then have i:"i = 0" 
    unfolding vars_term_list.simps by simp 
  then show ?case unfolding i vars_term_list.simps poss_list.simps varposs.simps by simp
next
  case (Fun f ts)
  let ?xs="(map vars_term_list ts)"
  let ?ys="(map2 (\<lambda>i. map ((#) i)) [0..<length ts] (map varposs_list ts))"
  from Fun(2) have 1:"i < length (concat ?xs)" unfolding vars_term_list.simps by simp
  have 2:"length ?ys = length ?xs" unfolding length_map by simp
  {fix i assume i:"i < length ?xs"
    then have *:"(map2 (\<lambda>x. map ((#) x)) [0..<length ts] (map varposs_list ts) ! i) = map ((#) i) (varposs_list (ts!i))" 
      unfolding length_map by simp
    with i have "length (?ys ! i) = length (?xs ! i)"
      unfolding * length_map length_varposs_list by simp
  }note l=this
  then obtain j k where j:"j < length ?xs" and k:"k < length (?xs ! j)" 
    and concat:"concat ?xs ! i = ?xs ! j ! k" "concat ?ys ! i = ?ys ! j ! k"
    using nth_concat_two_lists[OF 1 2] by blast 
  from Fun(1) j k have "Var (vars_term_list (ts!j) ! k) = (ts!j) |_ varposs_list (ts!j) ! k" 
    unfolding length_map by force 
  then have "Var (vars_term_list (Fun f ts) ! i) = (Fun f ts)|_(j#(varposs_list (ts!j) ! k))" 
    unfolding vars_term_list.simps concat(1) using j by auto 
  moreover have "j#(varposs_list (ts!j) ! k) = ((varposs_list (Fun f ts))!i)" proof-
    from k have "k < length (map2 (\<lambda>i. map ((#) i)) [0..<length ts] (map varposs_list ts) ! j)" 
      using l j by presburger 
    then show ?thesis 
      unfolding varposs_list_fun concat(2) using j unfolding length_map by simp
  qed
  ultimately show ?case
    by presburger  
qed

lemma varposs_list_map_vars_term:
  shows "varposs_list t = varposs_list (map_vars_term f t)"
proof(induct t)
  case (Fun g ts)
  then have IH:"map varposs_list ts = map varposs_list (map (map_vars_term f) ts)"
    by fastforce
  then show ?case unfolding map_vars_term_eq subst_apply_term.simps IH varposs_list_fun
    by force
qed (simp add:poss_list.simps)

lemma funposs_map_vars_term:
  "funposs (map_vars_term f t) = funposs t"
unfolding map_vars_term_eq proof(induct t)
  case (Fun g ts)
  {fix i assume "i < length ts"
    with Fun have "funposs (map (\<lambda>t. t \<cdot> (Var \<circ> f)) ts ! i) = funposs (ts!i)"
      by fastforce
    then have "{i # p |p. p \<in> funposs (map (\<lambda>t. t \<cdot> (Var \<circ> f)) ts ! i)} = {i # p |p. p \<in> funposs (ts ! i)}"
      by presburger
    }
  then show ?case unfolding funposs.simps subst_apply_term.simps length_map
    by auto
qed simp

lemma funposs_append_poss:
  assumes "p@q \<in> poss t" "q \<noteq> []"
  shows "p \<in> funposs t"
  by (meson assms is_Var_def poss_append_poss poss_is_Fun_funposs var_pos_maximal)

lemma funposs_append_poss':
  assumes "p@q \<in> funposs t"
  shows "p \<in> funposs t"
  by (metis append.right_neutral assms funposs_append_poss funposs_imp_poss)

lemma funposs_in_ctxt:
  assumes "q@p \<in> funposs (C\<langle>t\<rangle>)"
    and "hole_pos C = q"
  shows "p \<in> funposs t"
  by (metis Term.term.simps(4) assms(1) assms(2) funposs_fun_conv funposs_imp_poss hole_pos_poss hole_pos_poss_conv is_VarE poss_is_Fun_funposs subt_at_append subt_at_hole_pos)

lemma varposs_parallel:
  assumes "p \<in> varposs t" and "q \<in> varposs t" and "p \<noteq> q"
  shows "p \<bottom> q"
  using assms proof(induct t arbitrary:p q)
  case (Fun f ts)
  from Fun(2) obtain i p' where i:"i < length ts" "p' \<in> varposs (ts!i)" and p:"p = i#p'"
    using varposs_iff by fastforce 
  with Fun(3) obtain j q' where j:"j < length ts" "q' \<in> varposs (ts!j)" and q:"q = j#q'"
    using varposs_iff by fastforce 
  then show ?case proof(cases "i = j")
    case True
    from Fun(4) have "p' \<noteq> q'" 
      unfolding p q True by simp 
    with Fun(1) i j show ?thesis 
      unfolding True p q parallel_pos.simps using nth_mem by blast 
  next
    case False
    then show ?thesis unfolding p q
      by simp
  qed
qed simp

lemma not_below_poss:
  assumes "\<not> p \<le>\<^sub>p q"
    and "p \<in> poss t" and "q \<in> poss (replace_at t p s)"
  shows "q \<in> poss t"
  using assms by (metis order_pos.less_le parallel_poss_replace_at pos_cases replace_at_below_poss replace_at_ident)

section\<open>Linearity of terms\<close>
lemma var_in_linear_term_fun:
  assumes "linear_term (Fun f ts)"
    and "i < length ts" and "x \<in> vars_term (ts!i)" and "j < length ts \<and> j \<noteq> i"
  shows "x \<notin> vars_term (ts!j)"
proof-
  from assms(1) have "is_partition (map vars_term ts)"
    by simp
  with assms show ?thesis
    by (metis (mono_tags, lifting) IntI empty_iff is_partition_def le_less length_map not_less nth_map)
qed

lemma linear_term_var_vars_term_list:
  assumes "linear_term t"
  shows "vars_term_list t = var t"
proof-
 from assms have "\<forall>x. length (filter ((=) x) (vars_term_list t)) \<le> 1"
   by (meson linear_vars_term_list)
  then show ?thesis
    by (metis comp_def length_rev remdups_no_duplicates rev_filter rev_swap)
qed

lemma subt_at_linear:
  assumes "linear_term t" and "p \<in> poss t"
  shows "linear_term (t|_p)"
  using assms proof(induct p arbitrary:t)
  case (Cons i p)
  then obtain f ts where f:"t = Fun f ts" and i:"i < length ts" and p:"p \<in> poss (ts!i)"
    by (meson args_poss)
  with Cons(2) have "linear_term (ts!i)"
    by force
  then show ?case
    unfolding f subt_at.simps using Cons.hyps p by blast
qed simp

lemma linear_subterms_distinct_vars:
  assumes "linear_term t"
    and "p \<in> poss t" and "q \<in> poss t" and "p \<bottom> q"
  shows "vars_term (t|_p) \<inter> vars_term (t|_q) = {}"
  using assms proof(induct t arbitrary: p q)
  case (Fun f ts)
  from Fun(3,5) obtain i p' where i:"i < length ts" "p' \<in> poss (ts!i)" and p:"p = i#p'"
    by auto 
  with Fun(4,5) obtain j q' where j:"j < length ts" "q' \<in> poss (ts!j)" and q:"q = j#q'"
    by auto 
  then show ?case proof(cases "i=j")
    case True
    from Fun(5) have "p' \<bottom> q'" 
      unfolding p q True by simp
    with Fun(1,2) i j have "vars_term ((ts!j) |_ p') \<inter> vars_term ((ts!j) |_ q') = {}" 
      unfolding True by auto 
    then show ?thesis unfolding p q subt_at.simps True by simp
  next
    case False
    from i have "vars_term ((Fun f ts)|_p) \<subseteq> vars_term (ts!i)" 
      unfolding p subt_at.simps by (simp add: vars_term_subt_at) 
    moreover from j have "vars_term ((Fun f ts)|_q) \<subseteq> vars_term (ts!j)" 
      unfolding q subt_at.simps by (simp add: vars_term_subt_at) 
    ultimately show ?thesis using False Fun(2) i j
      by (meson disjoint_iff subsetD var_in_linear_term_fun)
  qed
qed simp

lemma linear_term_unique_vars:
  assumes "linear_term s"
    and "p \<in> poss s" and "s|_p = Var x"
    and "q \<in> poss s" and "s|_q = Var x"
  shows "p = q"
proof(rule ccontr)
  assume "p \<noteq> q"
  with assms(2-) obtain i j where ij:"i < length (varposs_list s)" "j < length (varposs_list s)" "i \<noteq> j" 
    "varposs_list s ! i = p" "varposs_list s ! j = q"
    by (metis varposs_iff varposs_in_varposs_list) 
  with assms(3,5) have "vars_term_list s ! i = vars_term_list s ! j"
    by (metis length_varposs_list term.inject(1) vars_term_list_varposs_list) 
  moreover from assms(1) have "distinct (vars_term_list s)"
    by (metis distinct_remdups distinct_rev linear_term_var_vars_term_list o_apply) 
  ultimately show False using ij(1,2,3)
    by (metis distinct_Ex1 length_varposs_list nth_mem) 
qed

lemma linear_term_ctxt:
  assumes "linear_term t"
    and "p \<in> poss t"
  shows "vars_ctxt (ctxt_of_pos_term p t) \<inter> vars_term (t|_p) = {}"
  using assms proof(induct p arbitrary:t)
  case (Cons i p)
  from Cons(3) obtain f ts where t:"t = Fun f ts" "i < length ts" "p \<in> poss (ts!i)"
    using args_poss by blast 
  with Cons(1,2) have IH:"vars_ctxt (ctxt_of_pos_term p (ts!i)) \<inter> vars_term ((ts!i) |_ p) = {}"
    by simp 
  {fix j assume j:"j < length ts" "j \<noteq> i"
    with Cons(2) have "vars_term (ts!j) \<inter> vars_term (ts!i |_p) = {}"
      unfolding t using var_in_linear_term_fun t(2,3) by (metis (no_types, opaque_lifting) Int_Un_distrib disjoint_iff sup_bot.neutr_eq_iff vars_ctxt_pos_term) 
  }
  then have "\<Union> {vars_term (ts ! j) |j. j < length ts \<and> j \<noteq> i} \<inter> vars_term (ts!i |_p) = {}"
    by blast 
  moreover have "(\<Union> (vars_term ` set (take i ts)) \<union> \<Union> (vars_term ` set (drop (Suc i) ts))) = 
                \<Union> {vars_term (ts ! j) |j. j < length ts \<and> j \<noteq> i}" 
    unfolding set_map[symmetric] take_map[symmetric] drop_map[symmetric] Union_Un_distrib[symmetric]
    using union_take_drop_list[where xs="(map vars_term ts)"] unfolding length_map using t(2) by auto
  ultimately show ?case unfolding t ctxt_of_pos_term.simps ctxt.set(4) subt_at.simps using IH
    by (metis (no_types, lifting) Un_empty inf_sup_distrib2)  
qed simp

section\<open>Results about substitutions and contexts.\<close>
lemma ctxt_comp_equals:
  assumes poss:"p \<in> poss s" "p \<in> poss t"
    and "ctxt_of_pos_term p s \<circ>\<^sub>c C = ctxt_of_pos_term p t \<circ>\<^sub>c D"
  shows "C = D"
  using assms proof(induct p arbitrary:s t)
  case (Cons i p)
  from Cons(2) obtain f ss where s:"s = Fun f ss" and i:"i < length ss" and p:"p \<in> poss (ss!i)"
    using args_poss by blast 
  from Cons(3) obtain g ts where t:"t = Fun g ts" and i:"i < length ts" and p':"p \<in> poss (ts!i)"
    using args_poss by blast 
  from Cons(1)[OF p p'] Cons(4) show ?case 
    unfolding s t ctxt_of_pos_term.simps ctxt_compose.simps by simp
qed simp

lemma ctxt_subst_comp_pos:
  assumes "q \<in> poss t" and "p \<in> poss (t \<cdot> \<tau>)"
    and "(ctxt_of_pos_term q t \<cdot>\<^sub>c \<sigma>) \<circ>\<^sub>c C = ctxt_of_pos_term p (t \<cdot> \<tau>)"
  shows "q \<le>\<^sub>p p"
  using assms by (metis hole_pos_ctxt_compose hole_pos_ctxt_of_pos_term hole_pos_subst less_eq_pos_simps(1)) 

lemma ctxt_apply_term_subst:
  assumes "linear_term t" and "i < length (vars_term_list t)"
    and "p = (varposs_list t)!i"
  shows "(ctxt_of_pos_term p (t \<cdot> \<sigma>))\<langle>s\<rangle> = t \<cdot> \<sigma>((vars_term_list t)!i := s)"
proof-
  from assms(2,3) have "t|_p = Var ((vars_term_list t)!i)"
    by (metis vars_term_list_varposs_list)
  with assms show ?thesis
    by (smt (verit, ccfv_threshold) filter_cong fun_upd_other fun_upd_same length_varposs_list linear_term_replace_in_subst nth_mem varposs_imp_poss varposs_set_varposs_list)
qed

lemma ctxt_subst_apply:
  assumes "p \<in> poss t" and "t|_p = Var x"
    and "linear_term t"
  shows "((ctxt_of_pos_term p t) \<cdot>\<^sub>c \<sigma>)\<langle>s\<rangle> = t \<cdot> \<sigma>(x := s)"
  using assms by (smt (verit) ctxt_of_pos_term_subst fun_upd_apply linear_term_replace_in_subst) 

lemma ctxt_of_pos_term_hole_subst:
  assumes "linear_term t" 
    and "i < length (varposs_list t)" and "p = varposs_list t ! i"
    and "\<forall>x \<in> vars_term t. x \<noteq> vars_term_list t !i \<longrightarrow> \<sigma> x = \<tau> x"
  shows "ctxt_of_pos_term p (t \<cdot> \<sigma>) = ctxt_of_pos_term p (t \<cdot> \<tau>)"
  using assms proof(induct p arbitrary: t i)
  case (Cons j p)
  from Cons(3,4) have "j#p \<in> varposs t"
    using nth_mem by force 
  then obtain f ts where ts:"j < length ts" "t = Fun f ts" "p \<in> varposs (ts!j)"
    by (metis args_poss subt_at.simps(2) varposs_iff) 
  then obtain i' where i':"i' < length (varposs_list (ts!j))" "p = varposs_list (ts!j)!i'"
    using varposs_in_varposs_list by blast 
  from Cons(3,4) have "Var (vars_term_list t ! i) = t|_(j#p)"
    by (metis length_varposs_list vars_term_list_varposs_list) 
  also have "... = (ts!j)|_p" 
    unfolding ts(2) by simp
  also have "... = Var (vars_term_list (ts!j) ! i')" 
    using i' by (simp add: length_varposs_list vars_term_list_varposs_list) 
  finally have *:"vars_term_list t ! i = vars_term_list (ts ! j) ! i'" by simp
  with Cons(5) have "\<forall>x\<in>vars_term (ts!j). x \<noteq> vars_term_list (ts!j) ! i' \<longrightarrow> \<sigma> x = \<tau> x" 
    unfolding ts(2) using ts(1) by auto 
  with Cons(2) i' ts have IH:"ctxt_of_pos_term p ((ts!j) \<cdot> \<sigma>) = ctxt_of_pos_term p ((ts!j) \<cdot> \<tau>)"  
    using Cons(1)[of "ts!j" i'] by (meson linear_term.simps(2) nth_mem) 
  {fix j' assume j':"j' < length ts" "j' \<noteq> j"
    with Cons(2) have "vars_term (ts ! j') \<inter> vars_term (ts!j) = {}" 
      unfolding ts(2) by (metis disjoint_iff ts(1) var_in_linear_term_fun)
    then have "\<forall>x \<in> vars_term (ts!j'). \<sigma> x = \<tau> x" 
      using Cons(5) j' * by (metis disjoint_iff i'(1) length_varposs_list nth_mem set_vars_term_list term.set_intros(4) ts(2)) 
    then have "(ts!j') \<cdot> \<sigma> = (ts!j') \<cdot> \<tau>"
      by (meson term_subst_eq)
  } note t'=this
  with ts(1) have "take j (map (\<lambda>t. t \<cdot> \<sigma>) ts) = take j (map (\<lambda>t. t \<cdot> \<tau>) ts)" 
    using take_equal_xs_ys by (metis (mono_tags, lifting) length_map nth_map) 
  moreover from t' ts(1) have "(drop (Suc j) (map (\<lambda>t. t \<cdot> \<sigma>) ts)) = (drop (Suc j) (map (\<lambda>t. t \<cdot> \<tau>) ts))"
    using drop_equal_xs_ys by (smt (verit) length_map nth_map)
  ultimately show ?case 
    unfolding ts(2) subst_apply_term.simps ctxt_of_pos_term.simps using IH by (simp add: ts(1)) 
qed simp

lemma ctxt_ctxt_apply:
  assumes "q \<in> poss t"
  shows "ctxt_of_pos_term q (ctxt_of_pos_term q t)\<langle>s\<rangle> = ctxt_of_pos_term q t"
  using assms proof(induct q arbitrary:t)
  case (Cons i q)
  from Cons(2) obtain f ts where t:"t = Fun f ts" and i:"i < length ts" and q:"q \<in> poss (ts!i)"
    using args_poss by blast 
  show ?case unfolding t ctxt_of_pos_term.simps ctxt_apply_term.simps using Cons(1)[OF q]
    by (simp add: i less_imp_le_nat nth_append_take)
qed simp

lemma ctxt_apply_ctxt_apply:
  assumes "p \<in> poss t"
  shows "(ctxt_of_pos_term (p@q) ((ctxt_of_pos_term p t) \<langle>s\<rangle>)) \<langle>u\<rangle> = (ctxt_of_pos_term p t)\<langle>(ctxt_of_pos_term q s) \<langle>u\<rangle>\<rangle>"
  by (metis assms ctxt_ctxt ctxt_of_pos_term_append hole_pos_ctxt_of_pos_term hole_pos_id_ctxt hole_pos_poss replace_at_subt_at)

lemma replace_at_append_subst:
  assumes "linear_term t"
    and "p \<in> poss t" "t|_p = Var x"
  shows "(ctxt_of_pos_term (p@q) (t \<cdot> \<sigma>)) \<langle>s\<rangle> = t \<cdot> \<sigma>(x := (ctxt_of_pos_term q (\<sigma> x)) \<langle>s\<rangle>)"
  using assms proof(induct p arbitrary:t)
  case (Cons i p)
  then obtain f ts where t:"t = Fun f ts" and i:"i < length ts" and p:"p \<in> poss (ts!i)"
    by (meson args_poss)
  from Cons(4) have x:"(ts!i)|_p = Var x" 
    unfolding t by simp
  from Cons(2) have lin:"linear_term (ts!i)"
    using i t by simp
  have IH:"(ctxt_of_pos_term (p@q) ((ts!i) \<cdot> \<sigma>)) \<langle>s\<rangle> = (ts!i) \<cdot> \<sigma>(x := (ctxt_of_pos_term q (\<sigma> x)) \<langle>s\<rangle>)"
    using Cons(1)[OF lin p x] .
  let ?\<sigma>="\<sigma>(x := (ctxt_of_pos_term q (\<sigma> x)) \<langle>s\<rangle>)"
  {fix j assume j:"j < length ts" "j \<noteq> i"
    from x have "x \<in> vars_term (ts!i)"
      by (metis p subsetD term.set_intros(3) vars_term_subt_at)
    then have "x \<notin> vars_term (ts!j)" 
      using j Cons(2) unfolding t by (meson i var_in_linear_term_fun) 
    then have "(ts!j) \<cdot> \<sigma> = (ts!j) \<cdot> ?\<sigma>"
      by (simp add: term_subst_eq_conv)
  } note sigma=this
  then have "take i (map (\<lambda>t. t \<cdot> \<sigma>) ts) = take i (map (\<lambda>t. t \<cdot> ?\<sigma>) ts)"
    by (simp add: i take_equal_xs_ys) 
  moreover from sigma have "drop (Suc i) (map (\<lambda>t. t \<cdot> \<sigma>) ts) = drop (Suc i) (map (\<lambda>t. t \<cdot> ?\<sigma>) ts)"
    by (simp add: drop_equal_xs_ys i)
  ultimately show ?case 
    unfolding t append_Cons subst_apply_term.simps ctxt_of_pos_term.simps ctxt_apply_term.simps nth_map[OF i] IH
    by (metis i id_take_nth_drop length_map nth_map)
qed simp

lemma replace_at_funposs_not_below:
  assumes "\<not> p \<le>\<^sub>p q"
    and "p \<in> poss t" and "q \<in> funposs (replace_at t p s)"
  shows "q \<in> funposs t"
  using assms by (metis ctxt_supt_id funposs_ctxt_apply_term hole_pos_ctxt_of_pos_term less_eq_pos_def) 

lemma not_elem_subst_of:
  assumes "x \<notin> set (map fst xs)"
  shows "(subst_of xs) x = Var x"
  using assms proof (induct xs)
  case (Cons y xs)
  then show ?case unfolding subst_of_simps
    by (metis Term.term.simps(17) insert_iff list.simps(15) list.simps(9) singletonD subst_compose subst_ident)
qed simp

lemma subst_of_eval:
  assumes "(x, t) \<in> set ss"
    and "\<forall>(y,s) \<in> set ss. (y = x \<longrightarrow> s = t)"
    and "set (map fst ss) \<inter> \<Union> ((vars_term \<circ> snd) ` set ss) = {}"
  shows "subst_of ss x = t"
  using assms proof(induct ss)
  case (Cons a ss)
  show ?case proof(cases "(x,t) \<in> set ss")
    case True
    from Cons(1)[OF True] Cons(3,4) have sub:"subst_of ss x = t"
      by (simp add: disjoint_iff) 
    from Cons(2,4) have "fst a \<notin> vars_term t"
      by fastforce 
    then show ?thesis 
      unfolding subst_of_simps subst_compose sub by simp 
  next
    case False
    then have "x \<notin> set (map fst ss)"
      using Cons(3) by auto
    then have sub:"subst_of ss x = Var x"
      by (meson not_elem_subst_of) 
    from Cons(2) False have "a = (x, t)"
      by simp 
    then show ?thesis 
      unfolding subst_of_simps subst_compose sub by simp 
  qed
qed simp

lemma substitution_subterm_at:
  assumes "\<forall>j < length (vars_term_list l). \<sigma> (vars_term_list l ! j) = s |_ (varposs_list l ! j)"
    and "\<exists>\<tau>. l \<cdot> \<tau> = s"
  shows "l \<cdot> \<sigma> = s"
  using assms proof(induct l arbitrary:s)
  case (Var x)
  then show ?case 
    unfolding vars_term_list.simps poss_list.simps varposs.simps subst_apply_term.simps by simp 
next
  case (Fun f ts)
  from Fun(3) obtain ss where s:"s = Fun f ss" and l:"length ts = length ss"
    by fastforce 
  {fix i assume i:"i < length ts"
    {fix j assume j:"j<length (vars_term_list (ts!i))" 
      let ?p="varposs_list (ts!i) ! j"
      let ?x="vars_term_list (ts!i) ! j"
      let ?k="sum_list (map (length \<circ> vars_term_list) (take i ts)) + j"
      from i j have x:"?x = vars_term_list (Fun f ts) ! ?k"
        unfolding vars_term_list.simps by (simp add: concat_nth take_map) 
      have p:"varposs_list (Fun f ts) ! ?k = i # ?p" proof-
        from i have i':"i < length (map2 (\<lambda>x. map ((#) x)) [0..<length ts] (map varposs_list ts))" 
          by simp 
        from i j have "j < length ((map varposs_list ts) ! i)" 
          using length_varposs_list by (metis (mono_tags, lifting) nth_map) 
        with i have j':"j < length (map2 (\<lambda>x. map ((#) x)) [0..<length ts] (map varposs_list ts) ! i)"
          by simp
        {fix l assume "l < length ts"
          then have "(map length (map2 (\<lambda>x. map ((#) x)) [0..<length ts] (map varposs_list ts)))!l = (map (length \<circ> vars_term_list) ts) ! l" 
            using length_varposs_list by simp
        }  
        then have "map length (map2 (\<lambda>x. map ((#) x)) [0..<length ts] (map varposs_list ts)) = map (length \<circ> vars_term_list) ts" 
          using nth_equalityI[where ys="map (length \<circ> vars_term_list) ts"] by simp
        with i have k:"sum_list (map length (take i (map2 (\<lambda>x. map ((#) x)) [0..<length ts] (map varposs_list ts)))) + j = ?k"
          by (metis take_map)
        then have "varposs_list (Fun f ts) ! ?k = (map2 (\<lambda>i. map ((#) i)) [0..<length ts] (map varposs_list ts))!i !j" 
          unfolding varposs_list_fun using concat_nth[OF i' j'] by presburger
        also have "... = (map ((#) i) (varposs_list (ts!i)))!j" using i by simp
        also have "... = i # ?p" using nth_map j length_varposs_list by metis
        ultimately show ?thesis by simp
      qed
      from i j have k:"?k < length (vars_term_list (Fun f ts))" 
        unfolding vars_term_list.simps by (metis concat_nth_length length_map map_map nth_map take_map) 
      from Fun(2) k have "\<sigma> ?x = (ss!i) |_ (varposs_list (ts!i) ! j)" 
        unfolding x s using p by simp
    }
    then have "\<forall>j<length (vars_term_list (ts!i)).\<sigma> (vars_term_list (ts!i) ! j) = (ss!i) |_ varposs_list (ts!i) ! j" 
      by simp
    moreover from Fun(3) have "\<exists>\<tau>. (ts!i) \<cdot> \<tau> = ss!i" 
      unfolding subst_apply_term.simps s using i l by (metis nth_map term.inject(2))
    ultimately have "(ts!i) \<cdot> \<sigma> = ss!i " 
      using i Fun(1) nth_mem by blast  
  }
  then show ?case unfolding subst_apply_term.simps s 
    using l by (simp add: map_nth_eq_conv)
qed

lemma vars_map_vars_term:
  "map f (vars_term_list t) = vars_term_list (map_vars_term f t)"
unfolding map_vars_term_eq proof(induct t)
  case (Fun g ts)
  then have "map (\<lambda>xs. map f xs)(map vars_term_list ts) = map vars_term_list (map (\<lambda>t. t \<cdot> (Var \<circ> f)) ts)"
    by fastforce
  then show ?case unfolding vars_term_list.simps subst_apply_term.simps map_map map_concat
    by presburger
qed (simp add: vars_term_list.simps)

lemma mk_subst_renamed:
  assumes "length (var t) = length xs"
  shows "t \<cdot> (mk_subst Var (zip (var t) xs)) = (map_vars_term ((#) c) t) \<cdot> (mk_subst Var (zip (var (map_vars_term ((#) c) t)) xs))" 
proof-
  {fix x assume "x \<in> vars_term t" 
    then obtain i where i:"x = (var t)!i" "i < length (var t)"
      by (metis in_set_conv_nth set_vars_term_list vars_term_list_var) 
    with assms have 1:"(mk_subst Var (zip (var t) xs)) x = xs!i" 
      using mk_subst_distinct by (metis comp_apply distinct_remdups distinct_rev) 
    have "var (map_vars_term ((#) c) t) = map ((#) c) (var t)" 
      unfolding vars_map_vars_term[symmetric] comp_apply
      by (smt (verit) distinct_conv_nth distinct_remdups length_map list.inject nth_map remdups_id_iff_distinct remdups_map_remdups rev_map)  
    with assms i have 2:"(mk_subst Var (zip (var (map_vars_term ((#) c) t)) xs)) (c#x) = xs!i"
      using mk_subst_distinct by (metis (no_types, lifting) comp_apply distinct_remdups distinct_rev length_map nth_map) 
    from 1 2 have "(mk_subst Var (zip (var t) xs)) x = (mk_subst Var (zip (var (map_vars_term ((#) c) t)) xs)) (c#x)"
      by presburger
  }
  then show ?thesis
    by (simp add: apply_subst_map_vars_term term_subst_eq_conv) 
qed

lemma mk_subst_same:
  assumes "length xs = length ts" "distinct xs"
  shows "map (mk_subst f (zip xs ts)) xs = ts"
  using assms  by (simp add: mk_subst_distinct map_nth_eq_conv)

section\<open>Lists of @{type option}.\<close>
lemma length_those:
  assumes "those xs = Some ys"
  shows "length xs = length ys"
  using assms proof(induction xs arbitrary:ys)
  case Nil
  then show ?case by simp
next
  case (Cons a xs)
  from Cons(2) obtain ys' where ys':"those xs = Some ys'"
    by (smt not_None_eq option.case_eq_if option.simps(8) those.simps(2)) 
  from Cons(2) obtain y where y:"Some y = a"
    by (metis option.case_eq_if option.exhaust_sel option.simps(3) those.simps(2)) 
  from y ys' have "those (Cons a xs) = Some (Cons y ys')"
    by auto
  then show ?case using Cons ys'
    by auto
qed

lemma those_not_none_x: "those xs = Some ys \<Longrightarrow> x \<in> set xs \<Longrightarrow> x \<noteq> None"
proof (induction xs arbitrary: x ys)
  case (Cons a xs)
  from Cons(2) have "a \<noteq> None" using option.simps(4) by fastforce 
  from this Cons(2) have "those xs \<noteq> None" by auto
  then show ?case using Cons(1,3) \<open>a \<noteq> None\<close> by auto
qed (simp)

lemma those_not_none_xs: "list_all (\<lambda>x. x \<noteq> None) xs \<Longrightarrow> those xs \<noteq> None"
  by (induction xs) auto

lemma those_some:
  assumes "length xs = length ys" "\<forall>i < length xs. xs!i = Some (ys!i)" 
  shows "those xs = Some ys"
  using assms by (induct rule:list_induct2) (simp, force)

lemma those_some2:
  assumes "those xs = Some ys"
  shows "\<forall>i < length xs. xs!i = Some (ys!i)"
proof-
  from assms have "length xs = length ys" by (simp add: length_those)
  then show ?thesis using assms proof(induction xs ys rule:list_induct2)
    case (Cons x xs y ys)
    from Cons(3) have "x \<noteq> None" by (metis list.set_intros(1) those_not_none_x)
    with Cons(3) have *:"x = Some y" by force
    with Cons(3) have "those xs = Some ys" by force
    with * Cons(2) show ?case by (simp add: nth_Cons') 
  qed simp
qed

lemma exists_some_list:
  assumes "\<forall>i < length xs. (\<exists>y. xs!i = Some y)"
  shows "\<exists> ys. (\<forall>i < length xs. xs!i = Some (ys!i)) \<and> length ys = length xs"
  by (metis assms length_map nth_map option.sel)

section\<open>match (t \<cdot> \<sigma>) t = Some \<sigma> whenever the domain of \<sigma> is a subset of the variables in t\<close>
text\<open>First some helper lemmas.\<close>
lemma decompose_fst:
  assumes "decompose (Fun f ss) t = Some us"
  shows "map fst us = ss"
proof-
  from assms obtain ts where t:"t = Fun f ts" 
     by (metis (no_types, lifting) decompose_def option.distinct(1) decompose_Some is_FunE old.prod.case term.case_eq_if)
  with assms have "length ss = length ts"
    by blast 
  with assms(1) t show ?thesis
    by auto 
qed

lemma decompose_vars_term:
  assumes "decompose (Fun f ss) t = Some us"
  shows "vars_term (Fun f ss) = (\<Union>(a, b) \<in> set us. vars_term a)"
proof-
  have "vars_term (Fun f ss) = (\<Union> s \<in> set ss. vars_term s)"
    by (meson Term.term.simps(18))
  also have "... = (\<Union> s \<in> set (map fst us). vars_term s)"
    using assms decompose_fst by metis 
  finally show ?thesis
    using image_image by auto 
qed

lemma match_term_list_domain:
  assumes "match_term_list P \<sigma> = Some \<tau>"
  shows "\<forall>x. x \<notin> (\<Union> (a, b) \<in> set P. vars_term a) \<and> \<sigma> x = None  \<longrightarrow> \<tau> x = None"
  using assms proof(induct P \<sigma> rule:match_term_list.induct)
  case (2 x t P \<sigma>)
  then show ?case
    by (metis (mono_tags, lifting) Sup_insert Un_iff case_prod_conv fun_upd_idem_iff fun_upd_triv fun_upd_twist image_insert list.simps(15) match_term_list.simps(2) option.simps(3) term.set_intros(3))
next
  case (3 f ss g ts P \<sigma>)
  from 3(2) obtain us where us:"decompose (Fun f ss) (Fun g ts) = Some us"
    using  match_term_list.simps(3) option.distinct(1) option.simps(4) by fastforce
  with 3(2) have *:"match_term_list (us @ P) \<sigma> = Some \<tau>" 
    by auto
  from us have "(\<Union> (a, b) \<in> set ((Fun f ss, Fun g ts) # P). vars_term a) = (\<Union> s \<in> set ss. vars_term s) \<union> (\<Union> (a, b) \<in> set P. vars_term a)"
    by simp
  also have "... = (\<Union> (a, b) \<in> set (us@P). vars_term a)" 
    using us by (metis (mono_tags, lifting) Term.term.simps(18) UN_Un decompose_vars_term set_append)
  finally show ?case 
    using  3(1) us * by metis
qed simp_all

lemma match_subst_domain:
  assumes "match a b = Some \<sigma>"
  shows "subst_domain \<sigma> \<subseteq> vars_term b"
proof-
  from assms have "\<forall>x. x \<notin> vars_term b \<longrightarrow> \<sigma> x = Var x"
    unfolding match_def match_list_def subst_of_map_def using match_term_list_domain by fastforce
  then show ?thesis
    using subst_domain_def by fastforce 
qed

lemma match_trivial: 
  assumes "subst_domain \<sigma> \<subseteq> vars_term t"
  shows "match (t \<cdot> \<sigma>) t = Some \<sigma>"
proof-
  obtain \<tau> where tau:"match (t \<cdot> \<sigma>) t = Some \<tau>" and 1:"(\<forall>x\<in>vars_term t. \<sigma> x = \<tau> x)"
    by (meson match_complete') 
  from assms have 2:"\<forall>x. x \<notin> vars_term t \<longrightarrow> \<sigma> x = Var x"
    by (meson notin_subst_domain_imp_Var subset_eq)
  from tau have 3:"\<forall>x. x \<notin> vars_term t \<longrightarrow> \<tau> x = Var x" 
    using match_subst_domain notin_subst_domain_imp_Var by fastforce  
  from 1 2 3 show ?thesis
    by (metis subst_term_eqI tau term_subst_eq)
qed

section\<open>Applying mk_subst when the variables involved form a partition.\<close>
lemma map2_zip: "set (map fst (concat (map2 zip xs ys))) \<subseteq> set (concat xs)"
proof
  fix x assume x:"x \<in> set (map fst (concat (map2 zip xs ys)))"
  let ?l="min (length xs) (length ys)"
  from x obtain i where i:"i < ?l" "x \<in> set (map fst (zip (xs!i) (ys!i)))"
    by (smt (verit) case_prod_conv in_set_conv_nth length_map length_zip min.strict_boundedE nth_concat_split nth_map nth_zip) 
  then have "x \<in> set (xs!i)"
    by (metis in_set_takeD map_fst_zip_take) 
  then show "x \<in> set (concat xs)"
    using i(1) by (metis concat_nth concat_nth_length in_set_conv_nth min.strict_boundedE) 
qed

lemma mk_subst_apply_simp:
  assumes "x \<notin> set (map fst xs)"
  shows "(mk_subst f (xs@ys)) x = (mk_subst f ys) x"
  using assms unfolding mk_subst_def map_of_append
  by (simp add: dom_map_of_conv_image_fst map_add_dom_app_simps(3)) 

lemma mk_subst_concat_Cons:
  assumes "x \<in> set (map fst ss)"
  shows "mk_subst f (concat (ss#ss')) x = mk_subst f ss x"
proof-
  from assms obtain y where "map_of ss x = Some y"
    by (metis list.set_map map_of_eq_None_iff not_None_eq) 
  then show ?thesis unfolding mk_subst_def concat.simps map_of_append
    by simp
qed

lemma mk_subst_not_mem':
  assumes x: "x \<notin> set (map fst ss)"
  shows "mk_subst f ss x = f x"
proof -
  have "map_of ss x = None"
    unfolding map_of_eq_None_iff using x by auto
  then show ?thesis unfolding mk_subst_def by auto
qed

lemma mk_subst_concat: 
  assumes "length ss = length ts"
    and "is_partition (map vars_term ts)"
  shows "\<forall>i < length ts. (ts!i) \<cdot> (mk_subst Var (zip (vars_term_list (ts!i)) (ss!i))) = (ts!i) \<cdot> (mk_subst Var (concat (map2 zip (map vars_term_list ts) ss)))"
proof-
  {fix i assume i:"i < length ts"
    {fix x assume x:"x \<in> vars_term (ts!i)"
      have "concat (map2 zip (map vars_term_list ts) ss) = concat (map2 zip (map vars_term_list (take i ts)) (take i ss)) @ 
                                                           concat (map2 zip (map vars_term_list (drop i ts)) (drop i ss))"
        by (metis append_take_drop_id concat_append drop_map drop_zip take_map take_zip)
      with i assms(1) have cc:"concat (map2 zip (map vars_term_list ts) ss) = concat (map2 zip (map vars_term_list (take i ts)) (take i ss)) @ 
                                                                  concat (zip (vars_term_list (ts!i)) (ss!i) # 
                                                                  (map2 zip (map vars_term_list (drop (Suc i) ts)) (drop (Suc i) ss)))"
        by (smt (z3) concat_append id_take_nth_drop length_map length_take list.map(2) map_append prod.simps(2) zip_Cons_Cons zip_append)
      {fix j assume "j < length ts" and "j \<noteq> i"
        with i x assms(2) have "x \<notin> vars_term (ts!j)" 
          unfolding is_partition_alt is_partition_alt_def length_map by auto
        then have "x \<notin> set (vars_term_list (ts!j))"
          by simp 
      } note part=this
      then have "x \<notin> set (concat (map vars_term_list (take i ts)))"
        by (smt (verit) in_set_conv_nth length_map length_take less_not_refl2 min.strict_boundedE nth_concat_split nth_map nth_take) 
      then have "x \<notin> set (map fst (concat (map2 zip (map vars_term_list (take i ts)) (take i ss))))" 
        using map2_zip in_mono by fastforce 
      then have subst:"(mk_subst Var (concat (map2 zip (map vars_term_list ts) ss))) x = (mk_subst Var (concat (zip (vars_term_list (ts!i)) (ss!i) # 
                                                                  (map2 zip (map vars_term_list (drop (Suc i) ts)) (drop (Suc i) ss))))) x"
        unfolding cc using mk_subst_apply_simp by metis
      have "mk_subst Var (zip (vars_term_list (ts ! i)) (ss ! i)) x = (mk_subst Var (concat (map2 zip (map vars_term_list ts) ss))) x"
      proof(cases "x \<in> set (map fst (zip (vars_term_list (ts!i)) (ss!i)))")
        case True
        then show ?thesis 
          using mk_subst_concat_Cons subst by metis
      next
        case False
        {fix j assume j:"j < length (drop (Suc i) ts)"
          then have "(drop (Suc i) ts)!j = ts!(Suc i + j)"
              using Suc_leI i nth_drop by blast 
            moreover from i j have "Suc i + j < length ts"
              by (metis add.commute length_drop less_diff_conv) 
            ultimately have "x \<notin> set (vars_term_list ((drop (Suc i) ts)!j))" 
              using part by (metis Suc_n_not_le_n le_add1) 
          }        
          then have "x \<notin> set (concat (map vars_term_list (drop (Suc i) ts)))" 
            by (smt (verit) in_set_conv_nth length_map length_take less_not_refl2 min.strict_boundedE nth_concat_split nth_map nth_take) 
          then have "x \<notin> set (map fst (concat (map2 zip (map vars_term_list (drop (Suc i) ts)) (drop (Suc i) ss))))" 
            using map2_zip in_mono by fastforce 
          with False have "x \<notin> set (map fst (concat (zip (vars_term_list (ts!i)) (ss!i) # (map2 zip (map vars_term_list (drop (Suc i) ts)) (drop (Suc i) ss)))))" 
            unfolding concat.simps by (metis Un_iff map_append set_append)
          with False show ?thesis 
            unfolding subst using mk_subst_not_mem' by metis
        qed
      }
      then have "ts ! i \<cdot> mk_subst Var (zip (vars_term_list (ts ! i)) (ss ! i)) = ts ! i \<cdot> (mk_subst Var (concat (map2 zip (map vars_term_list ts) ss)))"
        using term_subst_eq by blast
    }
  then show ?thesis by simp
qed

section\<open>lemmas about partitions\<close>
text\<open>needed for reasoning about unification algorithm\<close>
lemma is_partition_union:
  assumes "is_partition (x#y#xs)"
  shows "is_partition ((x\<union>y)#xs)"
  by (smt (z3) Sup_insert UnE UnI2 assms disjoint_iff is_partition_Cons list.set(2))

lemma is_partition_reorder:
  assumes "is_partition (xs @ x # ys)"
  shows "is_partition (x # xs @ ys)"
proof-
  from assms have "is_partition (xs @ [x])"
    by (metis append.simps is_partition_sublist)
  then have "\<forall>x' \<in> set xs. x \<inter> x' = {}"
    by (metis Int_commute append_Cons_nth_left in_set_conv_nth is_partition_def length_append_singleton lessI nth_append_length)
  then have "x \<inter> \<Union> (set xs) = {}"
    by blast 
  moreover from assms have "x \<inter> \<Union> (set ys) = {}"
    by (metis append_Nil2 is_partition_Cons is_partition_sublist) 
  moreover from assms have "is_partition (xs@ys)"
    by (metis append_Cons append_Nil append_Nil2 is_partition_sublist) 
  ultimately show ?thesis
    by (smt (z3) Int_commute UnE Union_disjoint is_partition_Cons set_append) 
qed

lemma is_partition_append:
  assumes "is_partition xs" and "is_partition zs" 
    and "\<forall>i < length xs. xs!i \<inter> \<Union> (set zs) = {}"
  shows "is_partition (xs@zs)"
  by (smt (verit, del_insts) add_diff_inverse_nat assms(1) assms(2) assms(3) disjoint_iff is_partition_alt is_partition_alt_def length_append mem_simps(9) nat_add_left_cancel_less nth_append nth_mem)

lemma is_partition_intersection: 
  assumes "is_partition (xs@ys)"
  shows "\<forall> i < length xs. xs!i \<inter> \<Union> (set ys) = {}"
  using assms proof(induct xs)
  case (Cons x xs)
  {fix i assume i:"i < length ys"
    then have "ys!i = ((x#xs)@ys) ! ((length xs) + 1 + i)"
      by simp 
    with i Cons(2) have "x \<inter> ys!i = {}" unfolding is_partition_alt is_partition_alt_def
      by (smt (verit, ccfv_SIG) Cons.prems Union_iff append.simps(2) append_assoc disjoint_iff in_set_conv_decomp is_partition_Cons nth_mem) 
  }
  then have "x \<inter> \<Union> (set ys) = {}" unfolding is_partition_alt is_partition_alt_def
    by (metis Union_disjoint in_set_conv_nth inf_commute)
  with Cons(1) show ?case
    by (metis Cons.prems append_Cons in_set_conv_nth is_partition_Cons set_ConsD)
qed simp

lemma is_partition_reorder_list: 
  assumes "is_partition (xs@ys@zs@us)" 
  shows "is_partition (xs@zs@ys@us)"
proof-
  have p1:"is_partition (xs@zs)"
    by (metis append_Nil assms is_partition_sublist)
  have p2:"is_partition (ys@us)"
    by (metis append.right_neutral assms is_partition_sublist) 
  {fix i assume i:"i < length xs"
    from assms have "is_partition (xs@ys@us)"
      by (metis append.assoc append_Nil append_Nil2 is_partition_sublist) 
    with i have "xs!i \<inter> \<Union> (set (ys@us)) = {}"
      using is_partition_intersection by blast 
  }moreover 
  {fix i assume i:"i < length zs"
    from assms have "is_partition (zs@us)"
      by (metis append_self_conv is_partition_sublist)
    with i have "zs!i \<inter> \<Union> (set us) = {}"
      using is_partition_intersection by blast 
    moreover have "zs!i \<inter> \<Union> (set ys) = {}" proof-
      from assms have "is_partition (ys@zs)"
        by (metis append_Nil is_partition_sublist) 
      with i show ?thesis
        using in_set_idx is_partition_intersection by fastforce
    qed
    ultimately have "zs!i \<inter> \<Union> (set (ys@us)) = {}"
      unfolding set_append by blast 
  }
  ultimately show ?thesis using is_partition_append[OF p1 p2]
    by (simp add: nth_append) 
qed

lemma distinct_is_partition:
  assumes "distinct xs"
  shows "is_partition (map (\<lambda>x. {x}) xs)"
  using assms proof(induct xs)
  case (Cons x xs)
  then show ?case unfolding list.map(2) is_partition_Cons by force
qed (simp add: is_partition_Nil)

lemma subst_of_partition:
  assumes "set substs = set res"
    and "is_partition ((map ((\<lambda>x. {x}) \<circ> fst) res) @ (map (vars_term \<circ> snd) res))"
  shows "subst_of substs = subst_of res"
proof-
  {fix x
    have "subst_of substs x = subst_of res x" proof(cases "x \<in> set (map fst res)")
      case True
      then obtain t where t:"(x,t) \<in> set res"
        by auto 
      moreover have unique:"\<forall>(y,s) \<in> set res. (y = x \<longrightarrow> s = t)" proof-
        {fix s assume s:"(x, s) \<in> set res" "s \<noteq> t"
          from assms(2) have "is_partition (map ((\<lambda>x. {x}) \<circ> fst) res)"
            by (metis append_Nil is_partition_sublist) 
          moreover from t obtain i where i:"res!i = (x,t)" "i < length res"
            by (metis in_set_idx) 
          moreover with s obtain j where j:"res!j = (x,s)" "j < length res" "i \<noteq> j"
            by (metis Pair_inject in_set_idx) 
          ultimately have False
            by (smt (verit, del_insts) IntI comp_apply fst_conv insert_absorb insert_not_empty is_partition_alt is_partition_alt_def length_map map_nth_eq_conv singletonI) 
        }
        then show ?thesis  by blast 
      qed
      moreover have distinct:"set (map fst res) \<inter> \<Union> ((vars_term \<circ> snd) ` set res) = {}" 
        using assms(2) by (smt (verit, best) disjoint_iff in_set_idx is_partition_intersection length_map list.set_map map_nth_eq_conv o_def singletonI) 
      ultimately have resx:"subst_of res x = t"
        by (simp add: subst_of_eval)
      from t have "(x,t) \<in> set substs" 
        using assms(1) by simp 
      moreover have "\<forall>(y,s) \<in> set substs. (y = x \<longrightarrow> s = t)" 
        using assms(1) unique by presburger  
      moreover have "set (map fst substs) \<inter> \<Union> ((vars_term \<circ> snd) ` set substs) = {}" 
        using assms(1) distinct by simp 
      ultimately have "subst_of substs x = t"
        by (simp add: subst_of_eval)
      with resx show ?thesis
        by simp  
    next
      case False
      with assms(1) show ?thesis by (simp add: not_elem_subst_of)
    qed
  }
  then show ?thesis
    by blast 
qed

section\<open>Unification of linear and variable disjoint terms\<close>
definition left_substs :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> ('v \<times> ('f, 'v) term) list"
  where "left_substs s t = (let filtered_vars = filter (\<lambda>(_, p). p \<in> poss t) (zip (vars_term_list s)(varposs_list s))
          in map (\<lambda>(x, p). (x, t|_p)) filtered_vars)"

definition right_substs :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> ('v \<times> ('f, 'v) term) list"
  where "right_substs s t = (let filtered_vars = filter (\<lambda>(_, q). q \<in> funposs s) (zip (vars_term_list t)(varposs_list t))
          in map (\<lambda>(y, q). (y, s|_q)) filtered_vars)"

lemma left_substs_imp_props:
  assumes "(x, u) \<in> set (left_substs s t)"
  shows "\<exists>p. p \<in> poss s \<and> s|_p = Var x \<and> p \<in> poss t \<and> t|_p = u"
proof-
  from assms obtain p where 1:"(x, p) \<in> set (zip (vars_term_list s)(varposs_list s))" and 2:"p \<in> poss t" "t|_p = u"
    unfolding left_substs_def Let_def using Pair_inject case_prodE filter_set in_set_idx length_map map_nth_eq_conv member_filter nth_mem old.prod.case by auto 
  from 1 have p:"p \<in> poss s"
    by (metis filter_set member_filter poss_list_sound set_zip_rightD) 
  from 1 obtain i where "i < length (zip (vars_term_list s)(varposs_list s))" and "(vars_term_list s)!i = x" and "(varposs_list s)!i = p"
    by (smt (z3) Pair_inject length_zip mem_Collect_eq set_zip)
  then have "s|_p = Var x"
    by (metis length_zip min_less_iff_conj vars_term_list_varposs_list)
  with 2 p show ?thesis
    by blast 
qed

lemma props_imp_left_substs:
  assumes "p \<in> poss s" and "s|_p = Var x" and "p \<in> poss t" and "t|_p = u"
  shows "(x, u) \<in> set (left_substs s t)"
proof-
  from assms obtain i where "(varposs_list s)!i = p" and "(vars_term_list s)!i = x" 
    by (metis nth_Cons_0 varposs_iff varposs_in_varposs_list vars_term_list_varposs_list length_varposs_list vars_term_list.simps(1))  
  then have "(x, p) \<in> set (zip (vars_term_list s)(varposs_list s))"
    by (metis assms(1) assms(2) fst_conv in_set_zip length_varposs_list prod.sel(2) term.inject(1) varposs_iff varposs_in_varposs_list vars_term_list_varposs_list)
  with assms(3) have "(x, p) \<in> set (filter (\<lambda>(_, p). p \<in> poss t) (zip (vars_term_list s) (varposs_list s)))"
    by simp 
  then show ?thesis unfolding left_substs_def Let_def assms(4)[symmetric] 
    by (smt (z3) case_prod_conv in_set_conv_nth length_map map_nth_eq_conv)
qed

lemma right_substs_imp_props:
  assumes "(x, u) \<in> set (right_substs s t)"
  shows "\<exists>q. q \<in> funposs s \<and> s|_q = u \<and> q \<in> poss t \<and> t|_q = Var x"
proof-
  from assms obtain q where 1:"(x, q) \<in> set (zip (vars_term_list t)(varposs_list t))" and 2:"q \<in> funposs s" "s|_q = u"
    unfolding right_substs_def Let_def using Pair_inject case_prodE filter_set in_set_idx length_map map_nth_eq_conv member_filter nth_mem old.prod.case by auto 
  from 1 have q:"q \<in> poss t"
    by (metis filter_set member_filter poss_list_sound set_zip_rightD) 
  from 1 obtain i where "i < length (zip (vars_term_list t)(varposs_list t))" and "(vars_term_list t)!i = x" and "(varposs_list t)!i = q"
    by (smt (z3) Pair_inject length_zip mem_Collect_eq set_zip)
  then have "t|_q = Var x"
    by (metis length_zip min_less_iff_conj vars_term_list_varposs_list)
  with 2 q show ?thesis 
    by blast 
qed

lemma props_imp_right_substs:
  assumes "q \<in> funposs s" and "s|_q = u" and "q \<in> poss t" and "t|_q = Var x"
  shows "(x, u) \<in> set (right_substs s t)"
proof-
  from assms obtain i where "(varposs_list t)!i = q" and "(vars_term_list t)!i = x" 
    by (metis nth_Cons_0 varposs_iff varposs_in_varposs_list length_varposs_list vars_term_list_varposs_list vars_term_list.simps(1))  
  then have "(x, q) \<in> set (zip (vars_term_list t)(varposs_list t))"
    by (metis assms(3) assms(4) fst_conv in_set_zip length_varposs_list prod.sel(2) term.inject(1) varposs_iff varposs_in_varposs_list vars_term_list_varposs_list) 
  with assms(1) have "(x, q) \<in> set (filter (\<lambda>(_, p). p \<in> funposs s) (zip (vars_term_list t) (varposs_list t)))"
    by simp 
  then show ?thesis unfolding right_substs_def Let_def assms(2)[symmetric] 
    by (smt (z3) case_prod_conv in_set_conv_nth length_map map_nth_eq_conv)
qed

lemma map_fst_left_substs: 
  "set (map fst (left_substs s t)) \<subseteq> vars_term s" 
unfolding left_substs_def using zip_fst by fastforce

lemma map_snd_left_substs: 
  assumes "t' \<in> set (map snd (left_substs s t))"
  shows "vars_term t' \<subseteq> vars_term t" 
proof-
  from assms obtain x where "(x, t') \<in> set (left_substs s t)"
    by force 
  then show ?thesis 
    using left_substs_imp_props by (metis vars_term_subt_at)
qed

lemma map_snd_right_substs: 
  assumes "t' \<in> set (map snd (right_substs s t))"
  shows "vars_term t' \<subseteq> vars_term s" 
proof-
  from assms obtain x where "(x, t') \<in> set (right_substs s t)"
    by force 
  then show ?thesis 
    using right_substs_imp_props by (metis funposs_imp_poss vars_term_subt_at)
qed

lemma distinct_map_fst_left_substs:
  assumes "linear_term t"
  shows "distinct (map fst (left_substs t s))"
proof-
  from assms have "distinct (vars_term_list t)"
    by (metis distinct_remdups distinct_rev linear_term_var_vars_term_list o_apply) 
  then have dist:"distinct (map fst (filter (\<lambda>(x, p). p \<in> poss s) (zip (vars_term_list t) (varposs_list t))))"
    by (simp add: distinct_map_filter length_varposs_list)
  have "map fst (left_substs t s) = (map fst (filter (\<lambda>(x, p). p \<in> poss s) (zip (vars_term_list t) (varposs_list t))))" 
    unfolding left_substs_def Let_def by auto 
  with dist show ?thesis
    by presburger 
qed

lemma distinct_map_fst_right_substs:
  assumes "linear_term t"
  shows "distinct (map fst (right_substs s t))"
proof-
  from assms have "distinct (vars_term_list t)"
    by (metis distinct_remdups distinct_rev linear_term_var_vars_term_list o_apply) 
  then have dist:"distinct (map fst (filter (\<lambda>(x, p). p \<in> funposs s) (zip (vars_term_list t) (varposs_list t))))"
    by (simp add: distinct_map_filter length_varposs_list)
  have "map fst (right_substs s t) = (map fst (filter (\<lambda>(x, p). p \<in> funposs s) (zip (vars_term_list t) (varposs_list t))))" 
    unfolding right_substs_def Let_def by auto 
  with dist show ?thesis
    by presburger 
qed

lemma is_partition_map_snd_left_substs:
  assumes "linear_term s" "linear_term t"
  shows "is_partition (map (vars_term \<circ> snd) (left_substs t s))"
proof-
  {fix i j assume j:"j < length (left_substs t s)" and i:"i < j" 
    from i j obtain x u where xu:"(x, u) = (left_substs t s)!i"
      by (metis surj_pair) 
    from i j obtain y v where yv:"(y, v) = (left_substs t s)!j"
      by (metis surj_pair) 
    from xu i j obtain p where p:"p \<in> poss t"  "t|_p = Var x"  "p \<in> poss s"  "s|_p = u" 
      using left_substs_imp_props by (metis Suc_lessD less_trans_Suc nth_mem) 
    from yv i j obtain q where q:"q \<in> poss t"  "t|_q = Var y"  "q \<in> poss s"  "s|_q = v" 
      using left_substs_imp_props by (metis nth_mem) 
    from assms(2) have "distinct (map fst (left_substs t s))" 
      using distinct_map_fst_left_substs by blast
    with xu yv i j have "x \<noteq> y"
      by (metis (mono_tags, lifting) Suc_lessD distinct_map eq_key_imp_eq_value less_trans_Suc nat_neq_iff nth_eq_iff_index_eq nth_mem)  
    with p(1,2) q(1,2) have "p \<bottom> q"
      by (metis term.inject(1) varposs_iff varposs_parallel) 
    with assms(1) p(3,4) q(3,4) have "vars_term (snd ((left_substs t s)!i)) \<inter> vars_term (snd ((left_substs t s)!j)) = {}"
      by (metis linear_subterms_distinct_vars snd_eqD xu yv) 
  }
  then show ?thesis unfolding is_partition_def map_map[symmetric] by auto  
qed

lemma is_partition_map_snd_right_substs:
  assumes "linear_term s" "linear_term t"
  shows "is_partition (map (vars_term \<circ> snd) (right_substs t s))"
proof-
  {fix i j assume j:"j < length (right_substs t s)" and i:"i < j" 
    from i j obtain x u where xu:"(x, u) = (right_substs t s)!i"
      by (metis surj_pair) 
    from i j obtain y v where yv:"(y, v) = (right_substs t s)!j"
      by (metis surj_pair) 
    from xu i j obtain p where p:"p \<in> poss s"  "s|_p = Var x"  "p \<in> funposs t"  "t|_p = u" 
      using right_substs_imp_props by (metis Suc_lessD less_trans_Suc nth_mem) 
    from yv i j obtain q where q:"q \<in> poss s"  "s|_q = Var y"  "q \<in> funposs t"  "t|_q = v" 
      using right_substs_imp_props by (metis nth_mem) 
    from assms(1) have "distinct (map fst (right_substs t s))" 
      using distinct_map_fst_right_substs by blast
    with xu yv i j have "x \<noteq> y"
      by (metis (mono_tags, lifting) Suc_lessD distinct_map eq_key_imp_eq_value less_trans_Suc nat_neq_iff nth_eq_iff_index_eq nth_mem)  
    with p(1,2) q(1,2) have "p \<bottom> q"
      by (metis term.inject(1) varposs_iff varposs_parallel) 
    with assms(2) p(3,4) q(3,4) have "vars_term (snd ((right_substs t s)!i)) \<inter> vars_term (snd ((right_substs t s)!j)) = {}"
      by (metis funposs_imp_poss linear_subterms_distinct_vars snd_eqD xu yv) 
  }
  then show ?thesis unfolding is_partition_def map_map[symmetric] by auto  
qed

lemma distinct_fst_lsubsts_snd_rsubsts:
  assumes "linear_term s"
  shows "(set (map fst (left_substs s t))) \<inter> \<Union> (set (map (vars_term \<circ> snd) (right_substs s t))) = {}"
proof-
  {fix x u assume "(x,u) \<in> set (left_substs s t)"
    then obtain p where p:"p \<in> poss s" "s|_p = Var x" "p \<in> poss t"  "t|_p = u"
      by (meson left_substs_imp_props) 
    {fix y v assume "(y,v) \<in> set (right_substs s t)"
      then obtain q where q:"q \<in> poss t" "t|_q = Var y" "q \<in> funposs s"  "s|_q = v"
        by (meson right_substs_imp_props) 
      with p have "p \<bottom> q"
        by (metis Term.term.simps(4) append.right_neutral funposs_fun_conv funposs_imp_poss parallel_pos prefix_pos_diff var_pos_maximal)
      with assms p(1,2) q(3,4) have "x \<notin> vars_term v"
        using funposs_imp_poss linear_subterms_distinct_vars by fastforce
    }
    then have "x \<notin> \<Union> (set (map (vars_term \<circ> snd) (right_substs s t)))"
      by fastforce
  }
  then show ?thesis by fastforce
qed

lemma distinct_fst_rsubsts_snd_lsubsts:
  assumes "linear_term t"
  shows "(set (map fst (right_substs s t))) \<inter> \<Union> (set (map (vars_term \<circ> snd) (left_substs s t))) = {}"
proof-
  {fix x u assume "(x,u) \<in> set (right_substs s t)"
    then obtain p where p:"p \<in> poss t" "t|_p = Var x" "p \<in> funposs s"  "s|_p = u"
      by (meson right_substs_imp_props) 
    {fix y v assume "(y,v) \<in> set (left_substs s t)"
      then obtain q where q:"q \<in> poss s" "s|_q = Var y" "q \<in> poss t"  "t|_q = v"
        by (meson left_substs_imp_props) 
      with p have "p \<bottom> q"
        by (metis Term.term.simps(4) append.right_neutral funposs_fun_conv funposs_imp_poss parallel_pos prefix_pos_diff var_pos_maximal)
      with assms p(1,2) q(3,4) have "x \<notin> vars_term v"
        using funposs_imp_poss linear_subterms_distinct_vars by fastforce
    }
    then have "x \<notin> \<Union> (set (map (vars_term \<circ> snd) (left_substs s t)))"
      by fastforce
  }
  then show ?thesis by fastforce
qed

text\<open>connection between left_substs and right_substs and decomposition of functions\<close>
lemma decompose_left_substs:
  assumes "decompose (Fun f ss) (Fun g ts) = Some ds"
  shows "set (left_substs (Fun f ss) (Fun g ts)) = (\<Union>e\<in>set ds. set (left_substs (fst e) (snd e)))" (is "?left = ?right") 
proof
  from assms have ds:"ds = zip ss ts" 
    using decompose_Some by auto
  show "?left \<subseteq> ?right" proof
    fix x t assume "(x,t) \<in> set (left_substs (Fun f ss) (Fun g ts))"
    then obtain p where 1:"p \<in> poss (Fun f ss)" and 2:"(Fun f ss)|_p = Var x" and 3:"p \<in> poss (Fun g ts)" and 4:"(Fun g ts)|_p = t"
      by (meson left_substs_imp_props) 
    from 1 2 obtain j p' where j1:"j < length ss" and "p = j#p'" and "p' \<in> poss (ss!j)" and "(ss!j)|_p' = Var x"
      by auto 
    moreover with 3 4 have j2:"j < length ts" and "p' \<in> poss (ts!j)" and "(ts!j)|_p' = t" 
      by auto
    ultimately have "(x,t) \<in> set (left_substs (ss!j) (ts!j))"
      by (meson props_imp_left_substs)
    moreover have "((ss!j),(ts!j)) \<in> set ds" 
      unfolding ds using j1 j2 by (metis length_zip min_less_iff_conj nth_mem nth_zip) 
    ultimately show "(x,t) \<in> (\<Union>e\<in>set ds. set (left_substs (fst e) (snd e)))"
      by force    
  qed
  show "?right \<subseteq> ?left" proof
    fix x t assume "(x,t) \<in> (\<Union>e\<in>set ds. set (left_substs (fst e) (snd e)))"
    then obtain j where j1:"j < length ss" and j2:"j < length ts" and "(x,t) \<in> set (left_substs (ss!j) (ts!j))" 
      unfolding ds by (metis (no_types, lifting) UN_E in_set_zip) 
    then obtain p where 1:"p \<in> poss (ss!j)" and 2:"(ss!j)|_p = Var x" and 3:"p \<in> poss (ts!j)" and 4:"(ts!j)|_p = t"
      by (meson left_substs_imp_props)
    then have "j#p \<in> poss (Fun f ss)" and "(Fun f ss)|_(j#p) = Var x" and "(j#p) \<in> poss (Fun g ts)" and "(Fun g ts)|_(j#p) = t" 
      using j1 j2 by auto
    then show "(x,t) \<in> set (left_substs (Fun f ss) (Fun g ts))"
      by (meson props_imp_left_substs)
  qed
qed 

lemma decompose_right_substs:
  assumes "decompose (Fun f ss) (Fun g ts) = Some ds"
  shows "set (right_substs (Fun f ss) (Fun g ts)) = (\<Union>e\<in>set ds. set (right_substs (fst e) (snd e)))" (is "?left = ?right") 
proof
  from assms have ds:"ds = zip ss ts" 
    using decompose_Some by auto
  show "?left \<subseteq> ?right" proof
    fix x t assume "(x,t) \<in> set (right_substs (Fun f ss) (Fun g ts))"
    then obtain q where 1:"q \<in> funposs (Fun f ss)" and 2:"(Fun f ss)|_q = t" and 3:"q \<in> poss (Fun g ts)" and 4:"(Fun g ts)|_q = Var x"
      by (meson right_substs_imp_props) 
    from 3 4 obtain j q' where j1:"j < length ts" and "q = j#q'" and "q' \<in> poss (ts!j)" and "(ts!j)|_q' = Var x" 
      by auto 
    moreover with 1 2 have j2:"j < length ss" and "q' \<in> funposs (ss!j)" and "(ss!j)|_q' = t" 
      by auto
    ultimately have "(x,t) \<in> set (right_substs (ss!j) (ts!j))"
      by (meson props_imp_right_substs)
    moreover have "((ss!j),(ts!j)) \<in> set ds" 
      unfolding ds using j1 j2 by (metis length_zip min_less_iff_conj nth_mem nth_zip) 
    ultimately show "(x,t) \<in> (\<Union>e\<in>set ds. set (right_substs (fst e) (snd e)))"
      by force    
  qed
  show "?right \<subseteq> ?left" proof
    fix x t assume "(x,t) \<in> (\<Union>e\<in>set ds. set (right_substs (fst e) (snd e)))"
    then obtain j where j1:"j < length ss" and j2:"j < length ts" and "(x,t) \<in> set (right_substs (ss!j) (ts!j))" 
      unfolding ds by (metis (no_types, lifting) UN_E in_set_zip) 
    then obtain q where 1:"q \<in> funposs (ss!j)" and 2:"(ss!j)|_q = t" and 3:"q \<in> poss (ts!j)" and 4:"(ts!j)|_q = Var x"
      by (meson right_substs_imp_props)
    then have "j#q \<in> funposs (Fun f ss)" and "(Fun f ss)|_(j#q) = t" and "(j#q) \<in> poss (Fun g ts)" and "(Fun g ts)|_(j#q) = Var x" 
      using j1 j2 by auto
    then show "(x,t) \<in> set (right_substs (Fun f ss) (Fun g ts))"
      by (meson props_imp_right_substs)
  qed
qed 

lemma unify_linear_var_disjoint_terms:
  assumes "unify es substs = Some res"
      and "\<forall>t \<in> set (map fst es) \<union> set (map snd es). linear_term t"
      and "is_partition (set (map fst substs)#(map vars_term (map fst es))@(map vars_term (map snd es)))"
    shows "set res = set substs \<union> (\<Union>e \<in> set es. (set (left_substs (fst e) (snd e))) \<union> (set (right_substs (fst e) (snd e))))" 
  using assms proof(induct arbitrary: res substs rule:unify.induct)
  case (2 f ss g ts E)
  from 2(2) obtain ds where ds':"decompose (Fun f ss) (Fun g ts) = Some ds" 
    unfolding unify.simps by fastforce 
  then have ds:"ds = zip ss ts" and l:"length ss = length ts"
    by fastforce+
  with 2(3) have "\<forall>t \<in> set (map fst ds). linear_term t" 
    using map_fst_zip by (metis (no_types, lifting) UnCI fst_conv linear_term.simps(2) list.set_intros(1) list.simps(9))
  moreover from 2(3) ds l have "\<forall>t \<in> set (map snd ds). linear_term t" 
    using map_snd_zip by (metis (no_types, lifting) UnCI linear_term.simps(2) list.set_intros(1) list.simps(9) snd_conv) 
  moreover from ds' 2(2) have "unify (ds @ E) substs = Some res"
    by auto 
  moreover have "is_partition (set (map fst substs) # map vars_term (map fst (ds @ E)) @ map vars_term (map snd (ds @ E)))"
  proof-
    let ?xs="set (map fst substs) # map (vars_term \<circ> fst) E @ map (vars_term \<circ> snd) E"
    let ?ys="map vars_term ss @ map vars_term ts"
    from 2(4) have *:"is_partition (set (map fst substs) # [vars_term (Fun f ss)] @ map (vars_term \<circ> fst) E @ [vars_term (Fun g ts)] @ map (vars_term \<circ> snd) E)"
      by force
    then have "is_partition [vars_term (Fun f ss), vars_term (Fun g ts)]"
      by (metis append.simps is_partition_sublist) 
    then have disj:"vars_term (Fun f ss) \<inter> vars_term (Fun g ts) = {}"
      by (simp add: is_partition_Cons) 
    from * have 1:"is_partition ?xs" 
      using is_partition_sublist by (metis (no_types, opaque_lifting) append_Cons append_Nil append_Nil2) 
    have 2:"is_partition ?ys" proof-
      from 2(3) have "is_partition (map vars_term ss)" by simp 
      moreover from 2(3) have "is_partition (map vars_term ts)" by simp 
      moreover from disj have "\<forall>i<length ss. vars_term (ss ! i) \<inter> \<Union> (set (map vars_term ts)) = {}" 
        unfolding term.set(4) by fastforce
      ultimately show ?thesis 
        using is_partition_append by force
    qed
    have 3:"\<forall>i < length ?xs. ?xs!i \<inter> \<Union> (set ?ys) = {}" proof-
      {fix i assume "i < length ?xs"
      then consider "i=0" | "i > 0 \<and> i < Suc (length E)" | "i > length E \<and> i < Suc (length E + length E)"
        using not_less_eq by fastforce
      then have "?xs!i \<inter> \<Union> (set ?ys) = {}" proof(cases)
        case 1
        from * have "is_partition (set (map fst substs) # [vars_term (Fun f ss)] @ [vars_term (Fun g ts)])"
          by (metis append_Cons append_Nil is_partition_sublist) 
        then show ?thesis
          by (simp add: "1" is_partition_Cons)
      next
        case 2
        then have "?xs!i=(map (vars_term \<circ> fst) E @ map (vars_term \<circ> snd) E)!(i-1)"
          by simp
        with 2 obtain j where j:"j < length E" "?xs!i = (map (vars_term \<circ> fst) E) ! j"
          by (metis One_nat_def Suc_pred length_map not_less_eq nth_append) 
        from * have "is_partition (map (vars_term \<circ> fst) E @ [vars_term (Fun f ss)] @ [vars_term (Fun g ts)])"
          by (metis (no_types, opaque_lifting) append_Nil append_eq_appendI is_partition_Cons is_partition_reorder_list is_partition_sublist) 
        then have "\<forall>j < length E. vars_term (fst (E!j)) \<inter> \<Union> (set ?ys) = {}" 
          using is_partition_intersection by fastforce
        then show ?thesis unfolding j(2) using j(1) by simp
      next
        case 3
        then have "?xs!i=(map (vars_term \<circ> fst) E @ map (vars_term \<circ> snd) E)!(i-1)"
          by simp
        with 3 obtain j where j:"j < length E" "?xs!i = (map (vars_term \<circ> snd) E) ! j" 
          by (metis (no_types, lifting) One_nat_def Suc_eq_plus1_left add_diff_inverse_nat length_map less_Suc0 less_nat_zero_code nat_add_left_cancel_less not_less_eq nth_append)
        from * have "is_partition ([vars_term (Fun f ss)] @ [vars_term (Fun g ts)] @ map (vars_term \<circ> snd) E)"
          by (metis append.right_neutral append_Nil is_partition_Cons is_partition_sublist)
        then have "is_partition (map (vars_term \<circ> snd) E @ [vars_term (Fun f ss)] @ [vars_term (Fun g ts)])" 
          using is_partition_reorder_list by (metis append.assoc append.right_neutral same_append_eq) 
        then have "\<forall>j < length E. vars_term (snd (E!j)) \<inter> \<Union> (set ?ys) = {}" 
          using is_partition_intersection by fastforce
        then show ?thesis unfolding j(2) using j(1) by simp
      qed}
    then show ?thesis by simp
    qed
    have "is_partition (set (map fst substs) # map vars_term (map fst (ds @ E)) @ map vars_term (map snd (ds @ E))) =
          is_partition (set (map fst substs) # map vars_term ss @ map (vars_term \<circ> fst) E @ map vars_term ts @ map (vars_term \<circ> snd) E)"
      unfolding ds using l by simp 
    also have "... = is_partition (set (map fst substs) # map (vars_term \<circ> fst) E @ map vars_term ss @ map vars_term ts @ map (vars_term \<circ> snd) E)" 
      using is_partition_reorder_list by (metis append.simps) 
    also have " ... = is_partition (?xs@?ys)" 
      using is_partition_reorder_list by (metis (no_types, opaque_lifting) append_Cons append_Nil append_Nil2)  
    finally show ?thesis using 1 2 3
      using is_partition_append by blast
  qed
  ultimately have IH:"set res = set substs \<union> (\<Union>e\<in>set (ds @ E). set (left_substs (fst e) (snd e)) \<union> set (right_substs (fst e) (snd e)))"
    using ds' 2(3) 2(1)[where res=res]
    by (metis (no_types, lifting) Un_iff insert_is_Un list.map(2) list.simps(15) map_append set_append)
  moreover from ds' have "set (left_substs (Fun f ss) (Fun g ts)) = (\<Union>e\<in>set ds. set (left_substs (fst e) (snd e)))"
    by (simp add: decompose_left_substs)
  moreover from ds' have "set (right_substs (Fun f ss) (Fun g ts)) = (\<Union>e\<in>set ds. set (right_substs (fst e) (snd e)))"
    by (simp add: decompose_right_substs)
  ultimately show ?case by auto
next
  case (3 x t E)
  from 3(5) have part:"is_partition ({x} # map vars_term (map fst E) @ map vars_term (t#(map snd E)))" 
    unfolding list.map(2) fst_conv snd_conv term.set(3)
    by (metis append_Cons append_Nil is_partition_sublist append_Nil2)
  then have "\<forall>e \<in> set E. x \<notin> vars_term (fst e)"
    by (simp add: is_partition_Cons) 
  moreover from part have "\<forall>i < length E. x \<notin> vars_term (snd (E!i))" 
    by (simp add: is_partition_Cons) 
  ultimately have subst:"subst_list (subst x t) E = E"
    by (simp add: map_nth_eq_conv subst_list_def)
  from part have x:"x \<notin> vars_term t" 
    by (simp add: is_partition_Cons)
  then have x2:"t \<noteq> Var x"
    by fastforce 
  with 3(3) x have unif:"unify E ((x, t) # substs) = Some res"
    using unify.simps(3) subst by auto
  from 3(5) have "is_partition (set (map fst substs) # {x} # map vars_term (map fst E) @ map vars_term (map snd E))" 
    unfolding list.map(2) fst_conv snd_conv term.set(3)
    by (metis append_Cons append_Nil is_partition_sublist append_Nil2)
  then have "is_partition (set (map fst ((x, t) # substs)) # map vars_term (map fst E) @ map vars_term (map snd E))"
    unfolding list.map fst_conv  using is_partition_union insert_is_Un by force
  with x2 x unif 3(2,4) have IH:"set res = set ((x, t) # substs) \<union> (\<Union>e\<in>set E. set (left_substs (fst e) (snd e)) \<union> set (right_substs (fst e) (snd e)))"
    unfolding subst by auto
  have *:"filter (\<lambda>(_, p). p \<in> poss t) (zip (vars_term_list (Var x))(varposs_list (Var x))) = [(x, [])]"
    by (simp add: poss_list.simps(1) vars_term_list.simps(1))  
  have left:"left_substs (Var x) t = [(x, t)]" 
    unfolding left_substs_def * by simp
  have **:"filter (\<lambda>(_, q). q \<in> funposs (Var x)) (zip (vars_term_list t) (varposs_list t)) = []"
    by simp 
  then have right:"right_substs (Var x) t = []" 
    unfolding right_substs_def ** by simp
  then show ?case using right left IH by force 
next
  case (4 f ts x E)
  from 4(4) have "is_partition (map vars_term ((Fun f ts)#(map fst E)) @ ({x} # map vars_term (map snd E)))" 
    unfolding list.map(2) fst_conv snd_conv term.set(3)
    by (metis append_Cons append_Nil is_partition_sublist append_Nil2)
  then have part:"is_partition ({x} # map vars_term (Fun f ts # map fst E) @ map vars_term (map snd E))"
    using is_partition_reorder by blast 
  then have "\<forall>e \<in> set E. x \<notin> vars_term (fst e)" 
    by (simp add: is_partition_reorder is_partition_Cons) 
  moreover from part have "\<forall>i < length E. x \<notin> vars_term (snd (E!i))" 
    by (simp add: is_partition_Cons) 
  ultimately have subst:"subst_list (subst x (Fun f ts)) E = E"
    by (simp add: map_nth_eq_conv subst_list_def)
  from part have x:"x \<notin> vars_term (Fun f ts)" 
    by (simp add: is_partition_Cons)
  with 4(2) x have unif:"unify E ((x, Fun f ts) # substs) = Some res"
    using unify.simps(3) subst by auto
  from 4(4) have "is_partition ({x} # set (map fst substs) # map vars_term (map fst E) @  map vars_term (map snd E))" 
    unfolding list.map(2) fst_conv snd_conv term.set(3) 
    by (metis is_partition_reorder append_Cons append_Nil is_partition_sublist append_Nil2)
  then have "is_partition (set (map fst ((x, Fun f ts) # substs)) # map vars_term (map fst E) @ map vars_term (map snd E))"
    unfolding list.map fst_conv  using is_partition_union insert_is_Un by force
  with x unif 4(1,3) have IH:"set res = set ((x, Fun f ts) # substs) \<union> (\<Union>e\<in>set E. set (left_substs (fst e) (snd e)) \<union> set (right_substs (fst e) (snd e)))"
    unfolding subst by auto
  have "\<forall>p \<in> set (varposs_list (Fun f ts)). p \<notin> {[]}" 
    unfolding varposs_set_varposs_list[symmetric] unfolding varposs.simps by blast 
  then have "\<forall>(_, p) \<in> set (zip (vars_term_list (Fun f ts)) (varposs_list (Fun f ts))). p \<notin> {[]}"
    using set_zip_rightD by fastforce
  then have *:"filter (\<lambda>(_, p). p \<in> poss (Var x)) (zip (vars_term_list (Fun f ts))(varposs_list (Fun f ts))) = []"
    unfolding poss.simps using filter_False by (smt (z3) case_prodE case_prod_conv) 
  have left:"left_substs (Fun f ts) (Var x) = []" 
    unfolding left_substs_def * by simp
  have **:"filter (\<lambda>(_, q). q \<in> funposs (Fun f ts)) (zip (vars_term_list (Var x)) (varposs_list (Var x))) = [(x,[])]"
    by (simp add: poss_list.simps(1) vars_term_list.simps(1)) 
  then have right:"right_substs (Fun f ts) (Var x) = [(x, Fun f ts)]" 
    unfolding right_substs_def ** by simp
  then show ?case using right left IH by force
qed auto

lemma mgu_linear_var_disjoint:
  assumes lin_s:"linear_term s" and lin_t:"linear_term t" 
    and unif:"unifiers {(s, t)} \<noteq> {}"
    and distinct:"distinct ((vars_term_list s) @ (vars_term_list t))"
  shows "mgu s t = Some (subst_of ((left_substs s t) @ (right_substs s t)))"
proof-
  let ?l_substs="left_substs s t"
  let ?r_substs="right_substs s t"
  have disj:"is_partition [vars_term s, vars_term t]"
    by (metis Sup_empty Sup_insert distinct distinct_append inf_bot_right is_partition_Cons is_partition_Nil list.set(1) list.set(2) set_vars_term_list sup_bot.right_neutral)
  from unif have "mgu s t \<noteq> None"
    by (meson mgu_complete) 
  then obtain us where "unify [(s, t)] [] = Some us" 
    unfolding mgu.simps by fastforce
  moreover with lin_s lin_t disj have "set us = set ?l_substs \<union> set ?r_substs"
    using unify_linear_var_disjoint_terms[where substs="[]" and es="[(s, t)]" and res=us] is_partition_Cons
    by auto 
  moreover have "is_partition (map ((\<lambda>x. {x}) \<circ> fst) (?l_substs @ ?r_substs) @ (map (vars_term \<circ> snd) (?l_substs @ ?r_substs)))" proof-
    have part1:"is_partition (map ((\<lambda>x. {x}) \<circ> fst) (?l_substs @ ?r_substs))" proof-
      have part1:"is_partition (map ((\<lambda>x. {x}) \<circ> fst) ?l_substs)" 
        using lin_s distinct_map_fst_left_substs distinct_is_partition by fastforce 
      have part2:"is_partition (map ((\<lambda>x. {x}) \<circ> fst) ?r_substs)" 
        using lin_t distinct_map_fst_right_substs distinct_is_partition by fastforce 
      {fix i assume i:"i < length ?l_substs"
        then have "(map ((\<lambda>x. {x}) \<circ> fst) ?l_substs)!i \<subseteq> vars_term s" 
          using map_fst_left_substs by (smt (verit, best) length_map nth_map nth_mem o_apply singletonD subsetI subset_code(1))
        moreover have "\<Union> (set (map ((\<lambda>x. {x}) \<circ> fst) ?r_substs)) \<subseteq> vars_term t" 
          unfolding right_substs_def using zip_fst by fastforce
        ultimately have "(map ((\<lambda>x. {x}) \<circ> fst) ?l_substs)!i \<inter> \<Union> (set (map ((\<lambda>x. {x}) \<circ> fst) ?r_substs)) = {}"
          using distinct by (metis Int_mono distinct_append set_vars_term_list subset_empty) 
      }
      then show ?thesis using is_partition_append[OF part1 part2] by simp
    qed
    have part2:"is_partition (map (vars_term \<circ> snd) (?l_substs @ ?r_substs))" proof-
      have part1:"is_partition (map (vars_term \<circ> snd) ?l_substs)" 
        using lin_s lin_t is_partition_map_snd_left_substs by blast 
      have part2:"is_partition (map (vars_term \<circ> snd) ?r_substs)" 
        using lin_s lin_t is_partition_map_snd_right_substs by blast 
      {fix i assume i:"i < length ?l_substs"
        then have "(map (vars_term \<circ> snd) ?l_substs)!i \<subseteq> vars_term t" 
          using map_snd_left_substs using nth_mem by fastforce
        moreover have "\<Union> (set (map (vars_term \<circ> snd) ?r_substs)) \<subseteq> vars_term s" 
          using map_snd_right_substs by auto 
        ultimately have "(map (vars_term \<circ> snd) ?l_substs)!i \<inter> \<Union> (set (map (vars_term \<circ> snd) ?r_substs)) = {}"
          using distinct using set_vars_term_list subset_empty by fastforce
      }
      then show ?thesis using is_partition_append[OF part1 part2] by simp
    qed
    {fix i assume i:"i < length (?l_substs @ ?r_substs)" 
      then obtain x where x:"{x} = (map ((\<lambda>x. {x}) \<circ> fst) (?l_substs @ ?r_substs))!i"
        by (metis (no_types, lifting) nth_map o_apply)
      with i consider "x \<in> set (map fst ?l_substs)" | "x \<in> set (map fst ?r_substs)"
        by (smt (verit, best) Un_iff length_map map_append nth_map nth_mem o_apply set_append singleton_inject)
      then have "x \<notin> \<Union> (set (map (vars_term \<circ> snd) (?l_substs @ ?r_substs)))" proof(cases)
        case 1
        have "\<Union> (set (map (vars_term \<circ> snd) ?l_substs)) \<subseteq> vars_term t" 
          using map_snd_left_substs by auto
        moreover from 1 have "x \<in> vars_term s"
          unfolding left_substs_def using zip_fst by fastforce
        ultimately have "x \<notin> \<Union> (set (map (vars_term \<circ> snd) ?l_substs))"
          by (smt (verit, best) Union_upper disj disjoint_iff is_partition_Cons list.set_intros(1) subsetD) 
        moreover from 1 have "x \<notin> \<Union> (set (map (vars_term \<circ> snd) ?r_substs))" 
          using distinct_fst_lsubsts_snd_rsubsts[OF lin_s] by blast
        ultimately show ?thesis by simp
      next
        case 2
        have "\<Union> (set (map (vars_term \<circ> snd) ?r_substs)) \<subseteq> vars_term s" 
          using map_snd_right_substs by auto
        moreover from 2 have "x \<in> vars_term t"
          unfolding right_substs_def using zip_fst by fastforce
        ultimately have "x \<notin> \<Union> (set (map (vars_term \<circ> snd) ?r_substs))"
          by (smt (verit, best) Union_upper disj disjoint_iff is_partition_Cons list.set_intros(1) subsetD) 
        moreover from 2 have "x \<notin> \<Union> (set (map (vars_term \<circ> snd) ?l_substs))" 
          using distinct_fst_rsubsts_snd_lsubsts[OF lin_t] by blast
        ultimately show ?thesis by simp
      qed
      then have "(map ((\<lambda>x. {x}) \<circ> fst) (?l_substs @ ?r_substs))!i \<inter> \<Union> (set (map (vars_term \<circ> snd) (?l_substs @ ?r_substs))) = {}"
        unfolding x[symmetric] by blast
    }
    then show ?thesis using is_partition_append[OF part1 part2] by simp 
  qed
  ultimately show ?thesis
    unfolding mgu.simps using subst_of_partition by (metis option.simps(5) set_append)
qed

end